from typing import List, Union

from netqasm.sdk.futures import Future, RegFuture
from netqasm.sdk.qubit import Qubit


def parity_meas(qubits: List[Qubit], bases: str) -> Union[Future, RegFuture, int]:
    """
    Performs a parity measurement on the provided qubits in the Pauli bases specified by 'bases'.
    `bases` should be a string with letters in 'IXYZ' and optionally start with '-'.
    If `bases` starts with '-', then the measurement outcome is flipped.
    The `basis` should have the same length as the number of qubits provided or +1 if starts with '-'.
    If more than one letter of 'bases' is not identity, then an ancilla qubit will be used, which is created
    using the connection of the first qubit.

    Parameters
    ----------
    qubits : List[:class:`~.sdk.qubit.Qubit`]
        The qubits to measure
    bases : str
        What parity meas to perform.

    Returns
    -------
    :class:~.sdk.futures.Future
        The measurement outcome
    """

    if bases.startswith("-"):
        negative = True
        bases = bases[1:]
    else:
        negative = False
    if not (len(qubits) == len(bases)):
        raise ValueError("Number of bases needs to be the number of qubits.")
    if not all([(B in "IXYZ") for B in bases]):
        raise ValueError("All elements of bases need to be in 'IXYZ'.")

    num_qubits = len(qubits)

    flip_basis = ["I"] * num_qubits
    non_identity_bases = []

    # declare outcome variable
    m: Union[Future, RegFuture, int]

    # Check if we need to flip the bases of the qubits
    for i in range(len(bases)):
        B = bases[i]
        if B == "X":
            flip_basis[i] = "H"
            non_identity_bases.append(i)
        elif B == "Y":
            flip_basis[i] = "K"
            non_identity_bases.append(i)
        elif B == "Z":
            non_identity_bases.append(i)
        else:
            pass

    if len(non_identity_bases) == 0:
        # Trivial measurement
        m = 0

    elif len(non_identity_bases) == 1:
        # Single_qubit measurement
        q_index = non_identity_bases[0]
        q = qubits[q_index]

        # Flip to correct basis
        if flip_basis[q_index] == "H":
            q.H()
        if flip_basis[q_index] == "K":
            q.K()

        # m = q.measure()
        m = q.measure(inplace=True)

        # # Flip the qubit back
        if flip_basis[q_index] == "H":
            q.H()
        if flip_basis[q_index] == "K":
            q.K()

    else:
        # Parity measurement, ancilla needed

        # Use the connection of the first qubit
        conn = qubits[0]._conn

        # Initialize ancilla qubit
        anc = Qubit(conn)

        # Flip to correct basis
        for i in range(len(bases)):
            if flip_basis[i] == "H":
                qubits[i].H()
            if flip_basis[i] == "K":
                qubits[i].K()

        # Transfer parity information to ancilla qubit
        for i in non_identity_bases:
            qubits[i].cnot(anc)

        # Measure ancilla qubit
        m = anc.measure()

        # Flip to correct basis
        for i in range(len(bases)):
            if flip_basis[i] == "H":
                qubits[i].H()
            if flip_basis[i] == "K":
                qubits[i].K()
    if negative:
        if not isinstance(m, Future):
            assert isinstance(m, int)
            m = 1 - m
        else:
            m.add(1, mod=2)
    return m



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum, auto
from typing import TYPE_CHECKING, Dict, List, Optional, Tuple, Union

from netqasm.lang.symbols import Symbols
from netqasm.lang.version import NETQASM_VERSION
from netqasm.util.string import rspaces

if TYPE_CHECKING:
    from netqasm.util import log

from netqasm.lang.operand import Label, Operand, Template

T_ProtoOperand = Union[int, Label, Operand]


class GenericInstr(Enum):
    # Allocation
    QALLOC = auto()
    # Initialization
    INIT = auto()
    INIT_ALL = auto()
    ARRAY = auto()
    SET = auto()
    # Memory
    STORE = auto()
    LOAD = auto()
    UNDEF = auto()
    LEA = auto()
    # Classical logic
    JMP = auto()
    BEZ = auto()
    BNZ = auto()
    BEQ = auto()
    BNE = auto()
    BLT = auto()
    BGE = auto()
    # Classical operations
    ADD = auto()
    SUB = auto()
    ADDM = auto()
    SUBM = auto()
    MUL = auto()
    DIV = auto()
    REM = auto()
    # Single-qubit gates
    X = auto()
    Y = auto()
    Z = auto()
    H = auto()
    S = auto()
    K = auto()
    T = auto()
    # Single-qubit rotations
    ROT_X = auto()
    ROT_Y = auto()
    ROT_Z = auto()
    # Two-qubit gates
    CNOT = auto()
    CPHASE = auto()
    # All-qubit gates
    ROT_X_ALL = auto()
    ROT_Y_ALL = auto()
    ROT_Z_ALL = auto()
    BICHROMATIC = auto()
    # Measurement
    MEAS = auto()
    MEAS_BASIS = auto()
    MEAS_ALL = auto()
    # Entanglement generation
    CREATE_EPR = auto()
    RECV_EPR = auto()
    # Waiting
    WAIT_ALL = auto()
    WAIT_ANY = auto()
    WAIT_SINGLE = auto()
    # Deallocation
    QFREE = auto()
    # Return
    RET_REG = auto()
    RET_ARR = auto()

    CROT_X = auto()
    CROT_Y = auto()
    CROT_Z = auto()

    # Move source qubit to target qubit (target is overwritten)
    MOV = auto()

    # Breakpoint
    BREAKPOINT = auto()


class BreakpointAction(Enum):
    NOP = 0
    DUMP_LOCAL_STATE = 1
    DUMP_GLOBAL_STATE = 2


class BreakpointRole(Enum):
    CREATE = 0
    RECEIVE = 1


def instruction_to_string(instr):
    if not isinstance(instr, GenericInstr):
        raise ValueError(f"Unknown instruction {instr}")
    return instr.name.lower()


def flip_branch_instr(instr: GenericInstr) -> GenericInstr:
    try:
        return {
            GenericInstr.BEQ: GenericInstr.BNE,
            GenericInstr.BNE: GenericInstr.BEQ,
            GenericInstr.BLT: GenericInstr.BGE,
            GenericInstr.BGE: GenericInstr.BLT,
            GenericInstr.BEZ: GenericInstr.BNZ,
            GenericInstr.BNZ: GenericInstr.BEZ,
        }[instr]
    except KeyError:
        raise ValueError(f"Not a branch instruction {instr}")


_STRING_TO_INSTRUCTION = {instruction_to_string(instr): instr for instr in GenericInstr}


def string_to_instruction(instr_str):
    instr = _STRING_TO_INSTRUCTION.get(instr_str)
    if instr is None:
        raise ValueError(f"Unknown instruction {instr_str}")
    return instr


def _get_lineo_str(lineno):
    if lineno is None:
        lineno = "()"
    else:
        lineno = f"({lineno})"
    return f"{rspaces(lineno, min_chars=5)} "


@dataclass
class ICmd:
    instruction: GenericInstr
    args: List[int] = None  # type: ignore
    operands: List[T_ProtoOperand] = None  # type: ignore
    lineno: Optional[log.HostLine] = None

    def __post_init__(self):
        if self.args is None:
            self.args = []
        if self.operands is None:
            self.operands = []

    def __str__(self):
        return self._build_str(show_lineno=False)

    @property
    def debug_str(self):
        return self._build_str(show_lineno=True)

    def _build_str(self, show_lineno=False):
        if len(self.args) == 0:
            args = ""
        else:
            args = Symbols.ARGS_DELIM.join(str(arg) for arg in self.args)
            args = Symbols.ARGS_BRACKETS[0] + args + Symbols.ARGS_BRACKETS[1]
        operands = " ".join(str(operand) for operand in self.operands)
        instr_name = instruction_to_string(self.instruction)
        if show_lineno:
            lineno_str = _get_lineo_str(self.lineno)
        else:
            lineno_str = ""
        return f"{lineno_str}{instr_name}{args} {operands}"


@dataclass
class BranchLabel:
    name: str
    lineno: Optional[log.HostLine] = None

    def _assert_types(self):
        assert isinstance(self.name, str)

    def __str__(self):
        return self._build_str(show_lineno=False)

    @property
    def debug_str(self):
        return self._build_str(show_lineno=True)

    def _build_str(self, show_lineno=False):
        if show_lineno:
            lineno_str = _get_lineo_str(self.lineno)
        else:
            lineno_str = ""
        return f"{lineno_str}{self.name}{Symbols.BRANCH_END}"


class ProtoSubroutine:
    """
    A `ProtoSubroutine` object represents a preliminary subroutine that consists
    of general 'commands' that might not yet be valid NetQASM instructions.
    These commands can include labels, or instructions with immediates that
    still need to be converted to registers.

    ProtoSubroutines can optionally have *arguments*, which are
    yet-to-be-defined variables that are used in one or more of the commands in
    the ProtoSubroutine. So, a ProtoSubroutine can be seen as a function which
    takes certain parameters (arguments). Concrete values for arguments can be
    given by instantiating (using the `instantiate` method).

    `ProtoSubroutine`s are currently only used by the sdk and the text parser
    (netqasm.parser.text). In both cases they are converted into `Subroutine`
    objects before given to other package components.
    """

    def __init__(
        self,
        commands: Optional[List[Union[ICmd, BranchLabel]]] = None,
        arguments: Optional[List[str]] = None,
        netqasm_version: Tuple[int, int] = NETQASM_VERSION,
        app_id: Optional[int] = None,
    ) -> None:
        self._netqasm_version: Tuple[int, int] = netqasm_version
        self._app_id: Optional[int] = app_id

        self._commands: List[Union[ICmd, BranchLabel]] = []
        if commands is not None:
            self.commands = commands

        self._arguments: List[str] = []
        if arguments is not None:
            self._arguments = arguments

        else:
            # figure out arguments by inspecting all commands
            for cmd in self.commands:
                if not isinstance(cmd, ICmd):
                    continue
                for op in cmd.operands:
                    if isinstance(op, Template):
                        self._arguments.append(op.name)

    @property
    def netqasm_version(self) -> Tuple[int, int]:
        return self._netqasm_version

    @property
    def app_id(self) -> Optional[int]:
        return self._app_id

    @property
    def commands(self) -> List[Union[ICmd, BranchLabel]]:
        return self._commands

    @commands.setter
    def commands(self, new_commands: List[Union[ICmd, BranchLabel]]) -> None:
        self._commands = new_commands

    @property
    def arguments(self) -> List[str]:
        return self._arguments

    def __str__(self):
        result = "ProtoSubroutine"

        if len(self.arguments) > 0:
            result += "(" + ",".join(arg_name for arg_name in self.arguments) + ")"
        result += "\n"
        result += f"  NetQASM version: {self.netqasm_version}\n"
        result += f"  App ID: {self.app_id}\n"

        result += " LN | HLN | CMD\n"
        for i, command in enumerate(self.commands):
            result += f"{rspaces(i)} {command.debug_str}\n"
        return result

    def instantiate(self, app_id: int, arguments: Dict[str, int]) -> None:
        commands: List[Union[ICmd, BranchLabel]] = []
        for cmd in self.commands:
            if not isinstance(cmd, ICmd):
                continue
            ops: List[T_ProtoOperand] = []
            for op in cmd.operands:
                if isinstance(op, Template):
                    ops.append(arguments[op.name])
                else:
                    ops.append(op)
            cmd.operands = ops
            commands.append(cmd)

        self.commands = commands
        self._app_id = app_id



from time import sleep


def as_completed(futures, names=None, sleep_time=0):
    futures = list(futures)
    if names is not None:
        names = list(names)
    while len(futures) > 0:
        for i, future in enumerate(futures):
            if future.done():
                futures.pop(i)
                if names is None:
                    yield future
                else:
                    name = names.pop(i)
                    yield future, name
        if sleep_time > 0:
            sleep(sleep_time)



from netqasm.qlink_compat import BellState
from netqasm.sdk.build_epr import EprMeasBasis, EprMeasureResult, basis_to_rotation
from netqasm.sdk.connection import DebugConnection
from netqasm.sdk.futures import Future


def create_measure_result(
    outcome: int,
    basis: EprMeasBasis,
    bell_state: BellState,
    expect_phi_plus: bool,
) -> EprMeasureResult:
    conn = DebugConnection("test")

    raw_outcome = Future(conn, 0, 0)
    raw_outcome._value = outcome

    remote_node = Future(conn, 1, 0)
    remote_node._value = 0

    duration = Future(conn, 2, 0)
    duration._value = 1000

    raw_bell_state = Future(conn, 3, 0)
    raw_bell_state._value = bell_state.value

    return EprMeasureResult(
        raw_measurement_outcome=raw_outcome,
        measurement_basis_local=basis_to_rotation(basis),
        measurement_basis_remote=basis_to_rotation(basis),
        post_process=expect_phi_plus,
        remote_node_id=remote_node,
        generation_duration=duration,
        raw_bell_state=raw_bell_state,
    )


def test_x_meas():
    for m in [0, 1]:
        result = create_measure_result(
            outcome=m,
            basis=EprMeasBasis.X,
            bell_state=BellState.PHI_PLUS,
            expect_phi_plus=True,
        )
        assert result.measurement_outcome == m

        result = create_measure_result(
            outcome=m,
            basis=EprMeasBasis.X,
            bell_state=BellState.PHI_MINUS,
            expect_phi_plus=True,
        )
        assert result.measurement_outcome == m ^ 1

        result = create_measure_result(
            outcome=m,
            basis=EprMeasBasis.X,
            bell_state=BellState.PSI_PLUS,
            expect_phi_plus=True,
        )
        assert result.measurement_outcome == m

        result = create_measure_result(
            outcome=m,
            basis=EprMeasBasis.X,
            bell_state=BellState.PSI_MINUS,
            expect_phi_plus=True,
        )
        assert result.measurement_outcome == m ^ 1


def test_y_meas():
    for m in [0, 1]:
        result = create_measure_result(
            outcome=m,
            basis=EprMeasBasis.Y,
            bell_state=BellState.PHI_PLUS,
            expect_phi_plus=True,
        )
        assert result.measurement_outcome == m

        result = create_measure_result(
            outcome=m,
            basis=EprMeasBasis.Y,
            bell_state=BellState.PHI_MINUS,
            expect_phi_plus=True,
        )
        assert result.measurement_outcome == m ^ 1

        result = create_measure_result(
            outcome=m,
            basis=EprMeasBasis.Y,
            bell_state=BellState.PSI_PLUS,
            expect_phi_plus=True,
        )
        assert result.measurement_outcome == m ^ 1

        result = create_measure_result(
            outcome=m,
            basis=EprMeasBasis.Y,
            bell_state=BellState.PSI_MINUS,
            expect_phi_plus=True,
        )
        assert result.measurement_outcome == m


def test_z_meas():
    for m in [0, 1]:
        result = create_measure_result(
            outcome=m,
            basis=EprMeasBasis.Z,
            bell_state=BellState.PHI_PLUS,
            expect_phi_plus=True,
        )
        assert result.measurement_outcome == m

        result = create_measure_result(
            outcome=m,
            basis=EprMeasBasis.Z,
            bell_state=BellState.PHI_MINUS,
            expect_phi_plus=True,
        )
        assert result.measurement_outcome == m

        result = create_measure_result(
            outcome=m,
            basis=EprMeasBasis.Z,
            bell_state=BellState.PSI_PLUS,
            expect_phi_plus=True,
        )
        assert result.measurement_outcome == m ^ 1

        result = create_measure_result(
            outcome=m,
            basis=EprMeasBasis.Z,
            bell_state=BellState.PSI_MINUS,
            expect_phi_plus=True,
        )
        assert result.measurement_outcome == m ^ 1


if __name__ == "__main__":
    test_x_meas()
    test_y_meas()
    test_z_meas()



"""Execution of application scripts without setting up a backend.

The `run_applications` function simply spawns a thread for each of the applications
given to it, and runs the Python script of each application.
The relevant quantum node controllers are expected to be setup elsewhere, e.g. as real
hardware that is connected to the machine that runs `run_applications`.
"""

from multiprocessing.pool import ThreadPool

from netqasm.logging.glob import get_netqasm_logger
from netqasm.logging.output import save_all_struct_loggers
from netqasm.util.thread import as_completed
from netqasm.util.yaml import dump_yaml

from .app_config import AppConfig
from .application import ApplicationInstance

logger = get_netqasm_logger()


def run_application(
    app_instance: ApplicationInstance,
    post_function=None,
    results_file=None,
    use_app_config=True,  # whether to give app_config as argument to app's main()
):
    programs = app_instance.app.programs

    with ThreadPool(len(programs) + 1) as executor:
        # Start the program threads
        program_futures = []
        for program in programs:
            inputs = app_instance.program_inputs[program.party]
            if use_app_config:
                app_cfg = AppConfig(
                    app_name=program.party,
                    node_name=app_instance.party_alloc[program.party],
                    main_func=program.entry,
                    log_config=app_instance.logging_cfg,
                    inputs=inputs,
                )
                inputs["app_config"] = app_cfg
            future = executor.apply_async(program.entry, kwds=inputs)
            program_futures.append(future)

        # Join the application threads and the backend
        program_names = [program.party for program in app_instance.app.programs]
        # NOTE: use app_<name> instead of prog_<name> for now for backward compatibility
        names = [f"app_{prog_name}" for prog_name in program_names]
        results = {}
        for future, name in as_completed(program_futures, names=names):
            results[name] = future.get()
        if results_file is not None:
            save_results(results=results, results_file=results_file)

    save_all_struct_loggers()


def save_results(results, results_file):
    dump_yaml(data=results, file_path=results_file)



"""Utility functions for Blind Quantum Computations"""


def measXY(q, angle):
    """Measure qubit `q` in the basis spanned by `|0> ± e^{i*angle} |1>`.
    This is equivalent to doing a Z-rotation of `angle`, then a Hadamard,
    and measuring in the Z-basis.
    Note: we use the convention that we rotate by +`angle` (not -`angle`).
    """
    q.rot_Z(angle=angle)
    q.H()
    return q.measure()


def teleport_state(epr_socket, theta):
    """Teleport a state Rz(theta)|+> to the server.
    The resulting state on the server's side is actually
    Rz(theta + m*pi) |+>, for the client's measurement outcome `m`.
    """
    epr = epr_socket.create_keep()[0]
    m = measXY(epr, theta)
    return m


def recv_teleported_state(epr_socket):
    """Let the client teleport a state to the server.
    The client will do a suitable measurement on her side.
    """
    return epr_socket.recv_keep()[0]


def send_meas_cmd(socket, phi):
    """Tell the server to measure the next qubit in angle `phi`.
    This effectively applies the operation H Rz(phi) on the logical input.
    """
    socket.send(str(phi))


def recv_meas_cmd(socket):
    """Receive the angle to measure the next qubit in."""
    return float(socket.recv())


def send_meas_outcome(socket, outcome):
    """Send the outcome (0 or 1) of the latest measurement to the client."""
    socket.send(str(outcome))


def recv_meas_outcome(socket):
    """Receive the measurement outcome (0 or 1) of the server's
    last measurement.
    """
    return int(socket.recv(maxsize=1))



"""Interface for classical broadcasting between Hosts.

This module contains the `BroadcastChannel` class which is a base for representing
classical broadcasting (sending classical messages to all other nodes in the network)
between Hosts.
"""
from __future__ import annotations

import abc
from timeit import default_timer as timer
from typing import TYPE_CHECKING, List, Optional, Tuple, Type

if TYPE_CHECKING:
    from . import socket as sck


class BroadcastChannel(abc.ABC):
    """Socket for sending messages to all nodes in the network (broadcasting).

    A BroadcastChannel can be used by Hosts to broadcast a message to all other nodes
    in the network. It is very similar to a Socket object, just without an explicit
    single remote node.

    A BroadcastChannel is a local object that each node (that wants to send and
    receive broadcast messages) must instantiate themselves.
    """

    def __init__(
        self,
        app_name: str,
        timeout: Optional[float] = None,
        use_callbacks: bool = False,
    ):
        """BroadcastChannel constructor.

        :param app_name: application/Host name of this channel's constructor
        :param timeout: maximum time to try and create this channel before aborting
        :param use_callbacks: whether to use the `recv_callback` and
            `conn_lost_callback` callback methods
        """
        pass

    @abc.abstractmethod
    def send(self, msg: str) -> None:
        """Broadcast a message to all remote nodes."""
        pass

    @abc.abstractmethod
    def recv(self, block: bool = True) -> Tuple[str, str]:
        """Receive a message that was broadcast and from whom."""
        pass

    def recv_callback(self, remote_app_name: str, msg: str) -> None:
        """This method gets called when a message is received.

        Subclass to define behaviour.

        NOTE: This only happens if `self.use_callbacks` is set to `True`.
        """
        pass

    def conn_lost_callback(self) -> None:
        """This method gets called when the connection is lost.

        Subclass to define behaviour.

        NOTE: This only happens if `self.use_callbacks` is set to `True`.
        """
        pass


class BroadcastChannelBySockets(BroadcastChannel):
    """Implementation of a BroadcastChannel using a Socket for each remote node.

    Technically this is a multicast channel since the receiving nodes must be
    explicitly listed. It simply uses one-to-one sockets for every remote node.
    """

    def __init__(self, app_name: str, remote_app_names: List[str], **kwargs):
        """BroadcastChannel constructor.

        :param app_name: application/Host name of self
        :param remote_app_names: list of receiving remote Hosts
        """
        self._sockets = {
            remote_app_name: self._socket_class(
                app_name=app_name, remote_app_name=remote_app_name, **kwargs
            )
            for remote_app_name in remote_app_names
        }

    @property
    @abc.abstractmethod
    def _socket_class(self) -> Type[sck.Socket]:
        pass

    def send(self, msg: str) -> None:
        """Broadcast a message to all remote nodes."""
        for socket in self._sockets.values():
            socket.send(msg=msg)

    def recv(
        self, block: bool = True, timeout: Optional[float] = None
    ) -> Tuple[str, str]:
        """Receive a message that was broadcast.

        Parameters
        ----------
        block : bool
            Whether to block for an available message
        timeout : float, optional
            Optionally use a timeout for trying to recv a message. Only used if `block=True`.

        Returns
        -------
        tuple
            (remote_node_name, msg)

        Raises
        ------
        RuntimeError
            If `block=False` and there is no available message
        """
        t_start = timer()
        while block:
            for remote_node_name, socket in self._sockets.items():
                try:
                    msg = socket.recv(block=False)
                except RuntimeError:
                    continue
                else:
                    return remote_node_name, msg
            if block and timeout is not None:
                t_now = timer()
                t_elapsed = t_now - t_start
                if t_elapsed > timeout:
                    raise TimeoutError(
                        "Timeout while trying to receive broadcasted message"
                    )
        raise RuntimeError("No message broadcasted")



import math

from netqasm.logging.output import get_new_app_logger
from netqasm.runtime.settings import Simulator, get_simulator
from netqasm.sdk import EPRSocket, Qubit
from netqasm.sdk.classical_communication.message import StructuredMessage
from netqasm.sdk.external import NetQASMConnection, Socket
from netqasm.sdk.toolbox import set_qubit_state


def main(app_config=None, phi=0.0, theta=0.0):
    # Inputs are coefficients of pi, e.g. phi=0.5 -> angle 0.5*pi
    phi *= math.pi
    theta *= math.pi

    log_config = app_config.log_config
    app_logger = get_new_app_logger(app_name="sender", log_config=log_config)

    # Create a socket to send classical information
    socket = Socket("sender", "receiver", log_config=log_config)

    # Create a EPR socket for entanglement generation
    epr_socket = EPRSocket("receiver")

    print("`sender` will start to teleport a qubit to `receiver`")

    # Initialize the connection to the backend
    sender = NetQASMConnection(
        app_name=app_config.app_name, log_config=log_config, epr_sockets=[epr_socket]
    )
    with sender:
        # Create a qubit to teleport
        q = Qubit(sender)
        set_qubit_state(q, phi, theta)

        # Create EPR pairs
        epr = epr_socket.create_keep()[0]

        # Teleport
        q.cnot(epr)
        q.H()
        m1 = q.measure()
        m2 = epr.measure()

    # Send the correction information
    m1, m2 = int(m1), int(m2)

    app_logger.log(f"m1 = {m1}")
    app_logger.log(f"m2 = {m2}")
    print(
        f"`sender` measured the following teleportation corrections: m1 = {m1}, m2 = {m2}"
    )
    print("`sender` will send the corrections to `receiver`")

    socket.send_structured(StructuredMessage("Corrections", (m1, m2)))

    # Send information about the original state to the other side,
    # just for visualizations purposes in QNE.
    if get_simulator() == Simulator.NETSQUID:
        socket.send_silent(str((phi, theta)))

    return {"m1": m1, "m2": m2}


if __name__ == "__main__":
    main()



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum, auto
from typing import List, Optional, Tuple

from netqasm.qlink_compat import BellState, EPRRole, EPRType, RandomBasis, TimeUnit
from netqasm.sdk.build_types import T_PostRoutine
from netqasm.sdk.futures import Array, Future, NoValueError


class EprMeasBasis(Enum):
    X = 0
    Y = auto()
    Z = auto()
    MX = auto()
    MY = auto()
    MZ = auto()


@dataclass
class EntRequestParams:
    remote_node_id: int
    epr_socket_id: int
    number: int
    post_routine: Optional[T_PostRoutine]
    sequential: bool
    time_unit: TimeUnit = TimeUnit.MICRO_SECONDS
    max_time: int = 0
    expect_phi_plus: bool = True
    min_fidelity_all_at_end: Optional[int] = None
    max_tries: Optional[int] = None
    random_basis_local: Optional[RandomBasis] = None
    random_basis_remote: Optional[RandomBasis] = None
    rotations_local: Tuple[int, int, int] = (0, 0, 0)
    rotations_remote: Tuple[int, int, int] = (0, 0, 0)


# Indices of Create Request arguments in serialized NetQASM array
SER_CREATE_IDX_TYPE = 0
SER_CREATE_IDX_NUMBER = 1
SER_CREATE_IDX_RANDOM_BASIS_LOCAL = 2
SER_CREATE_IDX_RANDOM_BASIS_REMOTE = 3
SER_CREATE_IDX_MINIMUM_FIDELITY = 4
SER_CREATE_IDX_TIME_UNIT = 5
SER_CREATE_IDX_MAX_TIME = 6
SER_CREATE_IDX_PRIORITY = 7
SER_CREATE_IDX_ATOMIC = 8
SER_CREATE_IDX_CONSECUTIVE = 9
SER_CREATE_IDX_PROBABILITY_DIST_LOCAL1 = 10
SER_CREATE_IDX_PROBABLIITY_DIST_LOCAL2 = 11
SER_CREATE_IDX_PROBABILITY_DIST_REMOTE1 = 12
SER_CREATE_IDX_PROBABLIITY_DIST_REMOTE2 = 13
SER_CREATE_IDX_ROTATION_X_LOCAL1 = 14
SER_CREATE_IDX_ROTATION_Y_LOCAL = 15
SER_CREATE_IDX_ROTATION_X_LOCAL2 = 16
SER_CREATE_IDX_ROTATION_X_REMOTE1 = 17
SER_CREATE_IDX_ROTATION_Y_REMOTE = 18
SER_CREATE_IDX_ROTATION_X_REMOTE2 = 19

# Length of NetQASM array for serialized Create Requests.
SER_CREATE_LEN = SER_CREATE_IDX_ROTATION_X_REMOTE2 + 1

# Indices of EPR Keep results in serialized NetQASM array
SER_RESPONSE_KEEP_IDX_TYPE = 0
SER_RESPONSE_KEEP_IDX_CREATE_ID = 1
SER_RESPONSE_KEEP_IDX_LOGICAL_QUBIT_ID = 2
SER_RESPONSE_KEEP_IDX_DIRECTONIALITY_FLAG = 3
SER_RESPONSE_KEEP_IDX_SEQUENCE_NUMBER = 4
SER_RESPONSE_KEEP_IDX_PURPOSE_ID = 5
SER_RESPONSE_KEEP_IDX_REMOTE_NODE_ID = 6
SER_RESPONSE_KEEP_IDX_GOODNESS = 7
SER_RESPONSE_KEEP_IDX_GOODNESS_TIME = 8
SER_RESPONSE_KEEP_IDX_BELL_STATE = 9

# Length of NetQASM array for EPR Keep results.
SER_RESPONSE_KEEP_LEN = SER_RESPONSE_KEEP_IDX_BELL_STATE + 1

# Indices of EPR Measure results in serialized NetQASM array
SER_RESPONSE_MEASURE_IDX_TYPE = 0
SER_RESPONSE_MEASURE_IDX_CREATE_ID = 1
SER_RESPONSE_MEASURE_IDX_MEASUREMENT_OUTCOME = 2
SER_RESPONSE_MEASURE_IDX_MEASUREMENT_BASIS = 3
SER_RESPONSE_MEASURE_IDX_DIRECTONIALITY_FLAG = 4
SER_RESPONSE_MEASURE_IDX_SEQUENCE_NUMBER = 5
SER_RESPONSE_MEASURE_IDX_PURPOSE_ID = 6
SER_RESPONSE_MEASURE_IDX_REMOTE_NODE_ID = 7
SER_RESPONSE_MEASURE_IDX_GOODNESS = 8
SER_RESPONSE_MEASURE_IDX_BELL_STATE = 9

# Length of NetQASM array for EPR Measure results.
SER_RESPONSE_MEASURE_LEN = SER_RESPONSE_MEASURE_IDX_BELL_STATE + 1


def serialize_request(tp: EPRType, params: EntRequestParams) -> List[Optional[int]]:
    """Convert an EntRequestParams object into a list of values that can be put
    in a NetQASM array."""
    array: List[Optional[int]] = [None for i in range(SER_CREATE_LEN)]

    array[SER_CREATE_IDX_TYPE] = tp.value
    array[SER_CREATE_IDX_NUMBER] = params.number

    # Only when max_time is 0, explicitly initialize the relavant array elements.
    # If it is max_time 0, these array element will be None.
    if params.max_time != 0:
        array[SER_CREATE_IDX_TIME_UNIT] = params.time_unit.value
        array[SER_CREATE_IDX_MAX_TIME] = params.max_time

    if tp == EPRType.M or tp == EPRType.R:
        # Only write when non-zero.
        if params.rotations_local != (0, 0, 0):
            array[SER_CREATE_IDX_ROTATION_X_LOCAL1] = params.rotations_local[0]
            array[SER_CREATE_IDX_ROTATION_Y_LOCAL] = params.rotations_local[1]
            array[SER_CREATE_IDX_ROTATION_X_LOCAL2] = params.rotations_local[2]
        if params.rotations_remote != (0, 0, 0):
            array[SER_CREATE_IDX_ROTATION_X_REMOTE1] = params.rotations_remote[0]
            array[SER_CREATE_IDX_ROTATION_Y_REMOTE] = params.rotations_remote[1]
            array[SER_CREATE_IDX_ROTATION_X_REMOTE2] = params.rotations_remote[2]
        if params.random_basis_local:
            array[SER_CREATE_IDX_RANDOM_BASIS_LOCAL] = params.random_basis_local.value
        if params.random_basis_remote:
            array[SER_CREATE_IDX_RANDOM_BASIS_REMOTE] = params.random_basis_remote.value

    return array


def deserialize_epr_keep_results(
    request: EntRequestParams, array: Array
) -> List[EprKeepResult]:
    """Convert values in a NetQASM array into EprKeepResult objects."""
    assert len(array) == request.number * SER_RESPONSE_KEEP_LEN
    results: List[EprKeepResult] = []
    for i in range(request.number):
        base = i * SER_RESPONSE_KEEP_LEN
        results.append(
            EprKeepResult(
                qubit_id=array.get_future_index(
                    base + SER_RESPONSE_KEEP_IDX_LOGICAL_QUBIT_ID
                ),
                remote_node_id=array.get_future_index(
                    base + SER_RESPONSE_KEEP_IDX_REMOTE_NODE_ID
                ),
                generation_duration=array.get_future_index(
                    base + SER_RESPONSE_KEEP_IDX_GOODNESS
                ),
                raw_bell_state=array.get_future_index(
                    base + SER_RESPONSE_KEEP_IDX_BELL_STATE
                ),
            )
        )
    return results


def deserialize_epr_measure_results(
    request: EntRequestParams, array: Array, role: EPRRole
) -> List[EprMeasureResult]:
    """Convert values in a NetQASM array into EprMeasureResult objects."""
    assert len(array) == request.number * SER_RESPONSE_MEASURE_LEN
    results: List[EprMeasureResult] = []
    for i in range(request.number):
        base = i * SER_RESPONSE_MEASURE_LEN
        results.append(
            EprMeasureResult(
                raw_measurement_outcome=array.get_future_index(
                    base + SER_RESPONSE_MEASURE_IDX_MEASUREMENT_OUTCOME
                ),
                measurement_basis_local=request.rotations_local,
                measurement_basis_remote=request.rotations_remote,
                post_process=(request.expect_phi_plus and role == EPRRole.RECV),
                remote_node_id=array.get_future_index(
                    base + SER_RESPONSE_MEASURE_IDX_REMOTE_NODE_ID
                ),
                generation_duration=array.get_future_index(
                    base + SER_RESPONSE_MEASURE_IDX_GOODNESS
                ),
                raw_bell_state=array.get_future_index(
                    base + SER_RESPONSE_MEASURE_IDX_BELL_STATE
                ),
            )
        )
    return results


@dataclass
class EprKeepResult:
    qubit_id: Future
    remote_node_id: Future
    generation_duration: Future
    raw_bell_state: Future

    @property
    def bell_state(self) -> BellState:
        assert self.raw_bell_state.value is not None
        return BellState(self.raw_bell_state.value)


def rotation_to_basis(rotations: Tuple[int, int, int]) -> Optional[EprMeasBasis]:
    if rotations == (0, 24, 0):
        return EprMeasBasis.X
    elif rotations == (8, 0, 0):
        return EprMeasBasis.Y
    elif rotations == (0, 0, 0):
        return EprMeasBasis.Z
    elif rotations == (0, 8, 0):
        return EprMeasBasis.MX
    elif rotations == (24, 0, 0):
        return EprMeasBasis.MY
    elif rotations == (16, 0, 0):
        return EprMeasBasis.MZ
    else:
        return None


def basis_to_rotation(basis: EprMeasBasis) -> Tuple[int, int, int]:
    if basis == EprMeasBasis.X:
        return (0, 24, 0)
    elif basis == EprMeasBasis.Y:
        return (8, 0, 0)
    elif basis == EprMeasBasis.Z:
        return (0, 0, 0)
    elif basis == EprMeasBasis.MX:
        return (0, 8, 0)
    elif basis == EprMeasBasis.MY:
        return (24, 0, 0)
    elif basis == EprMeasBasis.MZ:
        return (16, 0, 0)
    else:
        assert False, f"invalid EprMeasBasis {basis}"


@dataclass
class EprMeasureResult:
    raw_measurement_outcome: Future
    measurement_basis_local: Tuple[int, int, int]
    measurement_basis_remote: Tuple[int, int, int]
    post_process: bool
    remote_node_id: Future
    generation_duration: Future
    raw_bell_state: Future

    @property
    def measurement_outcome(self) -> int:
        """Get the measurement outcome, possibly post-processed.

        The outcome is post-processed only if the EPR create request indicated
        that the produced Bell state should be the Phi+ state, while the physically
        produced Bell state actually was another Bell state. In this case, the outcome
        is post-processed such that the statistics are *as if* the Phi+ state was
        produced and measured.

        Post-processing involves classical bit flips based on the physical Bell
        state produced and the measurement basis specified. If a Phi+ (or Phi_00)
        Bell state was actually produced physically, no post-processing is applied.
        If another Bell state was produced, a bit flip may be applied such that
        it looks like the Phi+ state was produced after all.

        If no post-processing is desired, use the `raw_measurement_outcome` instead.

        :return: post-processed measurement outcome
        """
        try:
            if not self.post_process:
                return int(self.raw_measurement_outcome)

            # else
            local = rotation_to_basis(self.measurement_basis_local)
            remote = rotation_to_basis(self.measurement_basis_remote)
            if local != remote:
                raise RuntimeError(
                    f"The local and remote measurement bases are not equal "
                    f"(local={local}, remote={remote}. Post-processed measurement "
                    f"outcome is not available. Use `raw_measurement_outcome` in "
                    f"combination with `measurement_basis_local` to interpret the "
                    f"outcome instead."
                )
            # We have local == remote.
            if local is None:  # not one of X, Y, Z
                raise RuntimeError(
                    f"The measurement basis is not one of X, Y, or Z (instead it is "
                    f"{local}. Post-processed measurement outcome is not available. "
                    f"Use `raw_measurement_outcome in combination with "
                    f"`measurement_basis_local` to interpet the outcome instead."
                )

            # This may raise a NoValueError
            m = int(self.raw_measurement_outcome)
            assert m == 0 or m == 1

            # Correct for Bell flips.
            if self.bell_state == BellState.PHI_MINUS:
                # correct for Z-gate applied to Phi+
                if local in [
                    EprMeasBasis.X,
                    EprMeasBasis.MX,
                    EprMeasBasis.Y,
                    EprMeasBasis.MY,
                ]:
                    m = m ^ 1
            elif self.bell_state == BellState.PSI_PLUS:
                # correct for X-gate applied to Phi+
                if local in [
                    EprMeasBasis.Y,
                    EprMeasBasis.MY,
                    EprMeasBasis.Z,
                    EprMeasBasis.MZ,
                ]:
                    m = m ^ 1
            elif self.bell_state == BellState.PSI_MINUS:
                # correct for X-gate and Z-gate applied to Phi+
                if local in [
                    EprMeasBasis.X,
                    EprMeasBasis.MX,
                    EprMeasBasis.Z,
                    EprMeasBasis.MZ,
                ]:
                    m = m ^ 1

            return m

        except NoValueError:
            raise ValueError(
                "The `measurement_outcome` property can only be used after "
                "the subroutine that produces this outcome has been flushed. "
                "This is because classical post-processing is done which can only "
                "happen when the outcome is available. "
                "To use the outcome as a future (without post-processing), use "
                "the `raw_measurement_outcome` attribute."
            )

    @property
    def bell_state(self) -> BellState:
        assert self.raw_bell_state.value is not None
        return BellState(self.raw_bell_state.value)



from netqasm.logging.glob import set_log_level
from netqasm.sdk.connection import DebugConnection
from netqasm.sdk.qubit import Qubit


def main(no_output=False):
    with DebugConnection("Alice") as alice:
        num = 30000

        res = alice.new_array(init_values=[0]).get_future_index(0)

        with alice.loop(num):
            q = Qubit(alice)
            m = q.measure()
            res.add(m)

        if no_output:
            print(f"binary:\n{alice.storage[0]}")


if __name__ == "__main__":
    set_log_level("INFO")
    main()



import numpy as np
import pytest

from netqasm.runtime.application import Program, default_app_instance
from netqasm.sdk import EPRSocket
from netqasm.sdk.external import NetQASMConnection, Socket, simulate_application
from netqasm.sdk.toolbox import create_ghz, get_angle_spec_from_float


@pytest.mark.parametrize(
    "angle, tol, expected_nds",
    [
        (np.pi, 1e-6, [(1, 0)]),
        (np.pi / 2, 1e-6, [(1, 1)]),
        (np.pi / 1024, 1e-6, [(1, 10)]),
        (np.pi * (1 + 1 / 2 + 1 / 4 + 1 / 16), 1e-6, [(16 + 8 + 4 + 1, 4)]),
        (np.pi * (1 + 1 / 2 + 1 / 4 + 1 / 16 + 1 / 1024), 1e-6, [(29, 4), (1, 10)]),
        (np.pi / 3, 1e-6, None),
        (1.5, 1e-6, None),
    ],
)
def test(angle, tol, expected_nds):
    print(angle)
    nds = get_angle_spec_from_float(angle=angle, tol=tol)
    if expected_nds is not None:
        assert nds == expected_nds
    print(nds)
    approx = sum(n * np.pi / 2**d for n, d, in nds)
    print(approx)
    assert np.abs(approx - angle) < tol


def run_node(node, down_node=None, up_node=None, do_corrections=False):
    # Setup EPR sockets, depending on the role of the node
    epr_sockets = []
    down_epr_socket = None
    down_socket = None
    up_epr_socket = None
    up_socket = None
    if down_node is not None:
        down_epr_socket = EPRSocket(down_node)
        epr_sockets.append(down_epr_socket)
        if do_corrections:
            down_socket = Socket(node, down_node)
    if up_node is not None:
        up_epr_socket = EPRSocket(up_node)
        epr_sockets.append(up_epr_socket)
        if do_corrections:
            up_socket = Socket(node, up_node)

    with NetQASMConnection(node, epr_sockets=epr_sockets):
        # Create a GHZ state with the other nodes
        q, corr = create_ghz(
            down_epr_socket=down_epr_socket,
            up_epr_socket=up_epr_socket,
            down_socket=down_socket,
            up_socket=up_socket,
            do_corrections=do_corrections,
        )
        m = q.measure()

    return (int(m), int(corr))


def _gen_create_ghz(num_nodes, do_corrections=False):

    # Setup the applications
    app_instance = default_app_instance(programs=[])
    for i in range(num_nodes):
        node = f"node{i}"
        if i == 0:
            down_node = None
        else:
            down_node = f"node{i - 1}"
        if i == num_nodes - 1:
            up_node = None
        else:
            up_node = f"node{i + 1}"
        app_instance.app.programs += [
            Program(
                party=node,
                entry=run_node,
                args=["node", "down_node", "up_node", "do_corrections"],
                results=[],
            )
        ]
        app_instance.program_inputs[node] = {
            "node": node,
            "down_node": down_node,
            "up_node": up_node,
            "do_corrections": do_corrections,
        }
        app_instance.party_alloc[node] = node

    # Run the applications
    outcomes = simulate_application(
        app_instance, use_app_config=False, enable_logging=False
    )[0]
    outcomes = {
        node: outcome for node, outcome in outcomes.items() if node != "backend"
    }
    print(outcomes)

    if do_corrections:
        corrected_outcomes = [m for (m, _) in outcomes.values()]
    else:
        corrected_outcomes = []
        # Check the outcomes
        correction = 0
        for i in range(num_nodes):
            node = f"app_node{i}"
            m, corr = outcomes[node]
            corrected_outcome = (m + correction) % 2
            corrected_outcomes.append(corrected_outcome)

            if 0 < i < num_nodes - 1:
                correction = (correction + corr) % 2

    print(corrected_outcomes)

    assert len(set(corrected_outcomes)) == 1


@pytest.mark.parametrize("do_corrections", [True, False])
@pytest.mark.parametrize("num_nodes", range(2, 6))
@pytest.mark.parametrize("i", range(5))  # Run 10 times
def test_create_ghz(do_corrections, num_nodes, i):
    _gen_create_ghz(num_nodes, do_corrections)



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))


# -- Project information -----------------------------------------------------

project = "netqasm"
copyright = "2021, QuTech"
author = "QuTech"

# The full version, including alpha/beta/rc tags
from importlib.metadata import version

release = version("netqasm")


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinx.ext.githubpages",
    "sphinx.ext.todo",
    "sphinx.ext.mathjax",
    "sphinx.ext.intersphinx",
    "sphinx.ext.doctest",
    "sphinx.ext.extlinks",
    "sphinx_autodoc_typehints",
    # 'sphinx_exercise',
]
napoleon_include_init_with_doc = True
autodoc_member_order = "bysource"
# autoclass_content = 'both'
always_document_param_types = True

# set_type_checking_flag = True

extlinks = {
    "netsquid": ("https://netsquid.org/%s", ""),
    "simulaqron": ("http://www.simulaqron.org/%s", ""),
    "squidasm": ("https://github.com/QuTech-Delft/squidasm/%s", ""),
    "network-layer": ("https://arxiv.org/abs/2010.02575/%s", ""),
    "repcode": ("https://en.wikipedia.org/wiki/Quantum_error_correction/%s", ""),
    "netqasm-paper": (
        "https://gitlab.tudelft.nl/qinc-wehner/netqasm/netqasm-paper/%s",
        "",
    ),
    # TODO add link to qnodeos
    "qnodeos": ("%s", ""),
}

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]



from dataclasses import dataclass
from typing import List, Union

import numpy as np

from netqasm.lang.operand import Immediate, Operand, Register, Template

from netqasm.util.quantum_gates import get_rotation_matrix
from . import base, core


@dataclass  # type: ignore
class AllQubitsRotationInstruction(base.ImmImmInstruction):
    @property
    def angle_num(self):
        return self.imm0

    @angle_num.setter
    def angle_num(self, new_val: Immediate):
        self.imm0 = new_val

    @property
    def angle_denom(self):
        return self.imm1

    @angle_denom.setter
    def angle_denom(self, new_val: Immediate):
        self.imm1 = new_val

    @classmethod
    def from_operands(cls, operands: List[Union[Operand, int]]):
        assert len(operands) == 2
        imm0, imm1 = operands

        if isinstance(imm0, int):
            imm0 = Immediate(value=imm0)
        elif isinstance(imm0, Immediate):
            pass
        else:
            assert isinstance(imm0, Template) or isinstance(imm0, Register)
        if isinstance(imm1, int):
            imm1 = Immediate(value=imm1)
        elif isinstance(imm1, Immediate):
            pass
        else:
            assert isinstance(imm1, Template) or isinstance(imm1, Register)
        # We allow imm0, imm1 to be Templates OR registers
        return cls(imm0=imm0, imm1=imm1)  # type: ignore


@dataclass
class AllQubitsInitInstruction(base.NoOperandInstruction):
    id: int = 42
    mnemonic: str = "init_all"


@dataclass
class AllQubitsMeasInstruction(base.RegAddrInstruction):
    id: int = 43
    mnemonic: str = "meas_all"


@dataclass
class RotZInstruction(core.RotationInstruction):
    id: int = 29
    mnemonic: str = "rot_z"

    def to_matrix(self) -> np.ndarray:
        axis = [0, 0, 1]
        angle = self.angle_num.value * np.pi / 2**self.angle_denom.value
        return get_rotation_matrix(axis, angle)


@dataclass
class AllQubitsRotXInstruction(AllQubitsRotationInstruction):
    id: int = 44
    mnemonic: str = "rot_x_all"


@dataclass
class AllQubitsRotYInstruction(AllQubitsRotationInstruction):
    id: int = 45
    mnemonic: str = "rot_y_all"


@dataclass
class AllQubitsRotZInstruction(AllQubitsRotationInstruction):
    id: int = 46
    mnemonic: str = "rot_z_all"


@dataclass
class BichromaticInstruction(base.ImmImmInstruction):
    id: int = 47
    mnemonic: str = "bichromatic"

    @property
    def angle_num(self):
        return self.imm0

    @angle_num.setter
    def angle_num(self, new_val: Immediate):
        self.imm0 = new_val

    @property
    def angle_denom(self):
        return self.imm1

    @angle_denom.setter
    def angle_denom(self, new_val: Immediate):
        self.imm1 = new_val

    @classmethod
    def from_operands(cls, operands: List[Union[Operand, int]]):
        assert len(operands) == 2
        imm0, imm1 = operands

        if isinstance(imm0, int):
            imm0 = Immediate(value=imm0)
        elif isinstance(imm0, Immediate):
            pass
        else:
            assert isinstance(imm0, Template) or isinstance(imm0, Register)
        if isinstance(imm1, int):
            imm1 = Immediate(value=imm1)
        elif isinstance(imm1, Immediate):
            pass
        else:
            assert isinstance(imm1, Template) or isinstance(imm1, Register)
        # We allow imm0, imm1 to be Templates OR registers
        return cls(imm0=imm0, imm1=imm1)  # type: ignore



"""Runtime interface for simulators.

This module contains the RuntimeManager class which can be used as a base class by
simulators to handle setting up backends and running applications.
"""

from abc import abstractmethod

from netqasm.runtime.application import ApplicationOutput


class NetworkInstance:
    pass


class NetworkConfig:
    pass


class ApplicationInstance:
    pass


class RuntimeManager:
    @abstractmethod
    def start_backend(self) -> None:
        raise NotImplementedError

    @abstractmethod
    def get_network(self) -> NetworkInstance:
        raise NotImplementedError

    @abstractmethod
    def set_network(self, cfg: NetworkConfig) -> None:
        raise NotImplementedError

    @abstractmethod
    def run_app(self, app: ApplicationInstance) -> ApplicationOutput:
        raise NotImplementedError



from netqasm.logging.glob import get_netqasm_logger
from netqasm.runtime.application import default_app_instance
from netqasm.sdk import Qubit
from netqasm.sdk.external import NetQASMConnection, simulate_application

logger = get_netqasm_logger()


def run_alice():
    num = 10
    with NetQASMConnection("Alice") as alice:
        for _ in range(num):
            q = Qubit(alice)
            q.H()
            m = q.measure(inplace=True)
            with m.if_eq(1):
                q.X()

            zero = q.measure()
            alice.flush()
            assert zero == 0


def test_measure_if_future():
    app_instance = default_app_instance(
        [
            ("Alice", run_alice),
        ]
    )
    simulate_application(app_instance, use_app_config=False, enable_logging=False)



import random

import numpy as np

from netqasm.examples.lib.bqc import recv_meas_outcome, send_meas_cmd, teleport_state
from netqasm.logging.glob import get_netqasm_logger
from netqasm.sdk import EPRSocket
from netqasm.sdk.external import NetQASMConnection, Socket

logger = get_netqasm_logger()


def main(app_config=None, num_iter=3, theta=None, phi=None, r=None):
    socket = Socket("client", "server", log_config=app_config.log_config)
    epr_socket = EPRSocket("server")

    client = NetQASMConnection(
        app_name=app_config.app_name,
        log_config=app_config.log_config,
        epr_sockets=[epr_socket],
    )

    num_qubits = num_iter + 1

    if theta is None:
        theta = [0 for _ in range(num_qubits)]
    if phi is None:
        phi = [random.uniform(0, 2 * np.pi) for _ in range(num_iter)]
    if r is None:
        r = [0 for _ in range(num_iter)]

    with client:
        # Teleport states q[0] to q[num_qubits - 1] to the Server.
        # The resulting state q[i] might have a phase `pi`,
        # depending on outcome m[i].
        m = [None] * num_qubits
        for i in range(num_qubits):
            m[i] = teleport_state(epr_socket, theta[i])
        client.flush()

        # Convert outcomes to integers to use them in calculations below.
        m = [int(m[i]) for i in range(num_qubits)]

        # delta[i] will hold the actual measurement angle sent to the Server.
        delta = [None] * num_iter
        # s[i] will hold the measurement outcome for qubit q[i].
        s = [None] * num_iter

        # For r and s, temporarily add two 0s at the end,
        # so that we can use indices -1 and -2 for convenience.
        s.extend([0, 0])
        r.extend([0, 0])

        # Main loop. For each iteration i, we let the Server measure q[i].
        # We want to measure with angle phi[i], but initial phases
        # (m[i] and theta[i]), as well as previous measurement outcomes s[j]
        # and secret key bits r[j] are required to be compensated for.
        # The actual angle we send to the Server is then called delta[i].
        for i in range(num_iter):
            delta[i] = pow(-1, s[i - 1] ^ r[i - 1]) * phi[i]
            delta[i] += (s[i - 2] ^ r[i - 2]) * np.pi
            delta[i] += r[i] * np.pi
            # we have q[i] = Rz(m[i]*pi + theta[i]), compensate for this:
            delta[i] -= theta[i]
            delta[i] -= m[i] * np.pi

            send_meas_cmd(socket, delta[i])
            s[i] = recv_meas_outcome(socket)

        # remove last 2 temporary 0s
        s = s[0:num_iter]
        r = r[0:num_iter]

        return {
            "delta": delta,
            "s": s,
            "m": m,
            "theta": theta,
            "phi": phi,
            "r": r,
        }



from netqasm.lang.encoding import COMMAND_BYTES, COMMANDS
from netqasm.lang.parsing import parse_text_subroutine


def test_command_length():
    for command_class in COMMANDS:
        length = len(bytes(command_class()))
        print(f"{command_class.__name__}: {len(bytes(command_class()))}")
        assert length == COMMAND_BYTES


def test_encode():
    subroutine = """
# NETQASM 0.0
# APPID 0

// Setup classical registers
set C1 1     // constant 1
set Q0 0     // qubit virtual address
set R0 0     // loop index
set R1 0     // counts 0s
set R2 0     // counts 1s
set R3 30000 // iterations

// Loop entry
LOOP:
beq R0 R3 EXIT

// Loop body
qalloc Q0
init Q0
x Q0
meas Q0 M0

// Count outcomes
bez M0 ZERO
// bnz M0 ZERO
add R2 R2 C1
jmp END
ZERO:
add R1 R1 C1
END:

qfree Q0
add R0 R0 C1

// Loop exit
jmp LOOP
EXIT:

ret_reg R1
ret_reg R2
"""

    subroutine = parse_text_subroutine(subroutine)
    print(subroutine)

    print("\nInstructions:")
    for cstruct in subroutine.cstructs:
        print(f"{cstruct.__class__.__name__}: {bytes(cstruct)}")

    print("\nFull subroutine:")
    print(bytes(subroutine))


def test_encode_substitution():
    subroutine = """
# NETQASM 0.0
# APPID 0
# DEFINE ms @0

// Setup classical registers
set Q0 0
set R0 0  // loop variable
array(10) $ms

// Loop entry
LOOP:
beq R0 10 EXIT

// Loop body
qalloc Q0
init Q0
h Q0
meas Q0 M0

// Store to array
store M0 $ms[R0]

qfree Q0
add R0 R0 1

// Loop exit
jmp LOOP
EXIT:
"""

    subroutine = parse_text_subroutine(subroutine)
    print(subroutine)

    print("\nInstructions:")
    for cstruct in subroutine.cstructs:
        print(f"{cstruct.__class__.__name__}: {bytes(cstruct)}")

    print("\nFull subroutine:")
    print(bytes(subroutine))


def test_encode_rotations():
    subroutine = """
# NETQASM 0.0
# APPID 0

set Q0 0
qalloc Q0
init Q0

// Perform rotations
rot_x Q0 1 1  // rotate by 1 * pi / 1 = pi
rot_x Q0 1 4  // rotate by 1 * pi / 4 = pi / 4
rot_y Q0 7 22  // rotate by 7 pi / 22

qfree Q0
"""

    subroutine = parse_text_subroutine(subroutine)
    print(subroutine)

    print("\nInstructions:")
    for cstruct in subroutine.cstructs:
        print(f"{cstruct.__class__.__name__}: {bytes(cstruct)}")

    print("\nFull subroutine:")
    print(bytes(subroutine))


if __name__ == "__main__":
    test_encode()
    test_encode_substitution()
    test_encode_rotations()



import os
import sys


class ProgressBar:
    def __init__(self, maxitr: int):
        self.maxitr: int = maxitr
        self.itr: int = 0
        try:
            self.cols = os.get_terminal_size().columns
        except (OSError, AttributeError):
            self.cols = 60
        print("")
        self.update()

    def increase(self) -> None:
        self.itr += 1
        self.update()

    def update(self) -> None:
        cols = self.cols - 8
        assert self.itr <= self.maxitr
        ratio = float(self.itr) / self.maxitr
        procent = int(ratio * 100)
        progress = "=" * int(cols * ratio)
        sys.stdout.write("\r")
        sys.stdout.write("[%*s] %d%%" % (-cols, progress, procent))
        sys.stdout.flush()
        pass

    def close(self) -> None:
        print("")



import logging
from typing import Optional, Union

NETQASM_LOGGER = "NetQASM"


def get_netqasm_logger(sub_logger: Optional[str] = None) -> logging.Logger:
    logger = logging.getLogger(NETQASM_LOGGER)
    if sub_logger is None:
        return logger
    if isinstance(sub_logger, str):
        return logger.getChild(sub_logger)
    else:
        raise TypeError(f"sub_logger should be None or str, not {type(sub_logger)}")

    return logger


def set_log_level(level: Union[int, str]) -> None:
    logger = get_netqasm_logger()
    logger.setLevel(level)


def get_log_level(effective: bool = True) -> int:
    if effective:
        return get_netqasm_logger().getEffectiveLevel()
    else:
        return get_netqasm_logger().level


def _setup_netqasm_logger() -> None:
    logger = get_netqasm_logger()
    formatter = logging.Formatter("%(levelname)s:%(name)s:%(message)s")
    syslog = logging.StreamHandler()
    syslog.setFormatter(formatter)
    logger.addHandler(syslog)
    logger.propagate = False


_setup_netqasm_logger()



import math
from enum import Enum, auto
from typing import List, Optional, Union

from netqasm.lang.ir import BranchLabel, GenericInstr, ICmd, ProtoSubroutine
from netqasm.logging.glob import get_netqasm_logger
from netqasm.sdk.build_types import NVHardwareConfig
from netqasm.sdk.connection import DebugConnection
from netqasm.sdk.constraint import ValueAtMostConstraint
from netqasm.sdk.epr_socket import EPRSocket
from netqasm.sdk.futures import RegFuture
from netqasm.sdk.qubit import Qubit, QubitMeasureBasis
from netqasm.sdk.transpile import NVSubroutineTranspiler

logger = get_netqasm_logger()


class PatternWildcard(Enum):
    ANY_ONE = auto()
    ANY_ZERO_OR_MORE = auto()
    BRANCH_LABEL = auto()


class PatternMatcher:
    def __init__(
        self,
        commands: List[Union[ICmd, BranchLabel]],
        pattern: List[Union[GenericInstr, PatternWildcard]],
    ) -> None:
        self._commands = commands
        self._pattern = pattern

        self._pat_len: int = len(self._pattern)
        self._pat_idx: int = 0
        self._cmd_idx: int = 0
        self._match_start: Optional[int] = None

    def _record_match(self) -> None:
        if self._match_start is None:
            self._match_start = self._cmd_idx
        self._pat_idx += 1

    def _reset_match(self) -> None:
        self._match_start = None
        self._pat_idx = 0

    def _match_any_one(self) -> None:
        logger.debug("\tmatching ANY_ONE")
        self._record_match()

    def _match_branch_label(self) -> None:
        cmd = self._commands[self._cmd_idx]
        if isinstance(cmd, BranchLabel):
            logger.debug("\tmatching BranchLabel")
            self._record_match()
        else:
            self._reset_match()

    def _match_instr(self) -> None:
        pat = self._pattern[self._pat_idx]
        assert isinstance(pat, GenericInstr)

        cmd = self._commands[self._cmd_idx]
        if isinstance(cmd, ICmd):
            if cmd.instruction == pat:
                logger.debug(f"\tmatching instr {pat}")
                self._record_match()
                return
        self._reset_match()

    def _match_any_zero_or_more(self) -> None:
        assert self._pat_idx > 0, "wildcard at start of pattern not allowed"
        assert (
            self._pat_idx + 1 < self._pat_len
        ), "wildcard at end of pattern not allowed"

        next_pat = self._pattern[self._pat_idx + 1]
        assert (
            isinstance(next_pat, GenericInstr)
            or next_pat == PatternWildcard.BRANCH_LABEL
        ), "wildcard directly after ANY_ZERO_OR_MORE not allowed"

        cmd = self._commands[self._cmd_idx]
        if next_pat == PatternWildcard.BRANCH_LABEL:
            if isinstance(cmd, BranchLabel):
                logger.debug("\tmatching BranchLabel after * wildcard")
                self._pat_idx += 2
        elif isinstance(next_pat, GenericInstr):
            if isinstance(cmd, ICmd):
                if cmd.instruction == next_pat:
                    logger.debug(f"\tmatching instr {next_pat} after * wildcard")
                    self._pat_idx += 2
        else:
            assert False

    def match(self) -> bool:
        logger.debug(f"trying to match {self._pattern}")
        while True:
            if self._pat_idx == self._pat_len:
                return True

            if self._cmd_idx == len(self._commands):
                return False

            curr_pat = self._pattern[self._pat_idx]
            logger.debug(f"curr pat: {curr_pat}")

            if curr_pat == PatternWildcard.ANY_ONE:
                self._match_any_one()
            elif curr_pat == PatternWildcard.BRANCH_LABEL:
                self._match_branch_label()
            elif isinstance(curr_pat, GenericInstr):
                self._match_instr()
            elif curr_pat == PatternWildcard.ANY_ZERO_OR_MORE:
                self._match_any_zero_or_more()
            else:
                assert False

            self._cmd_idx += 1


class ProtoSubroutineInspector:
    def __init__(self, subroutine: ProtoSubroutine) -> None:
        self._subroutine = subroutine

    def contains_instr(self, instr_type: GenericInstr) -> bool:
        for cmd in self._subroutine.commands:
            cmd_instr_type = cmd.instruction if isinstance(cmd, ICmd) else None
            if cmd_instr_type == instr_type:
                return True
        return False

    def find_instr(self, instr_type: GenericInstr) -> List[ICmd]:
        instrs: List[ICmd] = []
        for cmd in self._subroutine.commands:
            if not isinstance(cmd, ICmd):
                continue
            if cmd.instruction == instr_type:
                instrs.append(cmd)
        return instrs

    def match_pattern(
        self, pattern: List[Union[GenericInstr, PatternWildcard]]
    ) -> bool:
        return PatternMatcher(self._subroutine.commands, pattern).match()


def test_simple():
    with DebugConnection("conn") as conn:
        q1 = Qubit(conn)
        q2 = Qubit(conn)
        q1.H()
        q2.X()
        q1.X()
        q2.H()

        subroutine = conn.builder.subrt_pop_pending_subroutine()
        print(subroutine)

    inspector = ProtoSubroutineInspector(subroutine)
    assert inspector.contains_instr(GenericInstr.QALLOC)
    assert inspector.contains_instr(GenericInstr.SET)
    assert not inspector.contains_instr(GenericInstr.ROT_X)

    assert inspector.match_pattern(
        [
            GenericInstr.QALLOC,
            GenericInstr.INIT,
            GenericInstr.SET,
        ]
    )

    assert inspector.match_pattern(
        [
            GenericInstr.H,
            PatternWildcard.ANY_ONE,
            GenericInstr.X,
        ]
    )


def test_create_epr():
    DebugConnection.node_ids = {
        "Alice": 0,
        "Bob": 1,
    }

    epr_socket = EPRSocket("Bob")

    with DebugConnection("Alice", epr_sockets=[epr_socket]) as conn:
        epr = epr_socket.create()[0]

        epr.rot_Z(angle=math.pi)
        epr.H()

        _ = epr.measure(store_array=False)

        subroutine = conn.builder.subrt_pop_pending_subroutine()
        print(subroutine)

    inspector = ProtoSubroutineInspector(subroutine)

    assert inspector.match_pattern(
        [
            GenericInstr.ARRAY,
            PatternWildcard.ANY_ZERO_OR_MORE,
            GenericInstr.CREATE_EPR,
            GenericInstr.WAIT_ALL,
            PatternWildcard.ANY_ZERO_OR_MORE,
            GenericInstr.ROT_Z,
            PatternWildcard.ANY_ZERO_OR_MORE,
            GenericInstr.H,
            PatternWildcard.ANY_ZERO_OR_MORE,
            GenericInstr.RET_ARR,
            PatternWildcard.ANY_ZERO_OR_MORE,
            GenericInstr.RET_REG,
        ]
    )


def test_branching():
    with DebugConnection("Alice") as conn:

        def if_true(conn):
            q = Qubit(conn)
            _ = q.measure()

        conn.if_eq(42, 42, if_true)

        q2 = Qubit(conn)
        m2 = q2.measure()
        with m2.if_ne(0):
            _ = Qubit(conn)

        with m2.if_ez():
            _ = Qubit(conn)

        subroutine = conn.builder.subrt_pop_pending_subroutine()
        print(subroutine)

    inspector = ProtoSubroutineInspector(subroutine)

    assert inspector.match_pattern(
        [
            GenericInstr.BNE,
            PatternWildcard.ANY_ZERO_OR_MORE,
            PatternWildcard.BRANCH_LABEL,
        ]
    )

    assert inspector.match_pattern(
        [
            GenericInstr.MEAS,
            GenericInstr.QFREE,
            GenericInstr.STORE,
            GenericInstr.LOAD,
            GenericInstr.BEQ,
            PatternWildcard.ANY_ZERO_OR_MORE,
            PatternWildcard.BRANCH_LABEL,
        ]
    )


def test_loop_context():
    with DebugConnection("conn") as conn:
        q = Qubit(conn)
        with conn.loop(2):
            q.H()

        subroutine = conn.builder.subrt_pop_pending_subroutine()
        print(subroutine)

    inspector = ProtoSubroutineInspector(subroutine)
    assert inspector.contains_instr(GenericInstr.QALLOC)
    assert inspector.contains_instr(GenericInstr.SET)
    assert not inspector.contains_instr(GenericInstr.ROT_X)

    assert inspector.match_pattern(
        [
            PatternWildcard.BRANCH_LABEL,
            GenericInstr.BEQ,
            PatternWildcard.ANY_ZERO_OR_MORE,
            GenericInstr.JMP,
            PatternWildcard.BRANCH_LABEL,
        ]
    )


def test_looping():
    with DebugConnection("Alice") as conn:

        def body(conn: DebugConnection, reg: RegFuture):
            q = Qubit(conn)
            _ = q.measure()

        conn.loop_body(body, 42, loop_register="C9")

        subroutine = conn.builder.subrt_pop_pending_subroutine()
        print(subroutine)

    inspector = ProtoSubroutineInspector(subroutine)

    assert inspector.match_pattern(
        [
            GenericInstr.SET,
            PatternWildcard.BRANCH_LABEL,
            GenericInstr.BEQ,
            PatternWildcard.ANY_ZERO_OR_MORE,
            GenericInstr.ADD,
            GenericInstr.JMP,
            PatternWildcard.BRANCH_LABEL,
        ]
    )


def test_futures():
    with DebugConnection("Alice") as conn:

        q = Qubit(conn)
        m = q.measure()
        with m.if_ne(0):
            _ = Qubit(conn)
        with m.if_ez():
            _ = Qubit(conn)

        subroutine = conn.builder.subrt_pop_pending_subroutine()
        print(subroutine)

    inspector = ProtoSubroutineInspector(subroutine)

    assert inspector.match_pattern(
        [
            GenericInstr.MEAS,
            GenericInstr.QFREE,
            GenericInstr.STORE,
            GenericInstr.LOAD,
            GenericInstr.BEQ,
            PatternWildcard.ANY_ZERO_OR_MORE,
            PatternWildcard.BRANCH_LABEL,
            GenericInstr.LOAD,
            GenericInstr.BNZ,
            PatternWildcard.ANY_ZERO_OR_MORE,
            PatternWildcard.BRANCH_LABEL,
        ]
    )


def test_nested():
    with DebugConnection("Alice") as conn:

        q = Qubit(conn)
        m = q.measure()
        with m.if_eq(0):
            with m.if_eq(1):
                _ = Qubit(conn)

        with conn.loop(2):
            with conn.loop(3):
                _ = Qubit(conn)

        with conn.loop(2):
            with m.if_eq(0):
                _ = Qubit(conn)

        with m.if_eq(0):
            with conn.loop(2):
                _ = Qubit(conn)

        subroutine = conn.builder.subrt_pop_pending_subroutine()
        print(subroutine)

    inspector = ProtoSubroutineInspector(subroutine)

    assert inspector.match_pattern(
        [
            GenericInstr.BNE,
            PatternWildcard.ANY_ZERO_OR_MORE,
            GenericInstr.BNE,
            PatternWildcard.ANY_ZERO_OR_MORE,
            PatternWildcard.BRANCH_LABEL,  # IF_EXIT
            PatternWildcard.BRANCH_LABEL,  # IF_EXIT1
            PatternWildcard.ANY_ZERO_OR_MORE,
            PatternWildcard.BRANCH_LABEL,  # LOOP1
            GenericInstr.BEQ,
            PatternWildcard.ANY_ZERO_OR_MORE,
            PatternWildcard.BRANCH_LABEL,  # LOOP
            GenericInstr.BEQ,
            PatternWildcard.ANY_ZERO_OR_MORE,
            GenericInstr.JMP,
            PatternWildcard.BRANCH_LABEL,  # LOOP_EXIT
            PatternWildcard.ANY_ZERO_OR_MORE,
            GenericInstr.JMP,
            PatternWildcard.BRANCH_LABEL,  # LOOP_EXIT1
            PatternWildcard.ANY_ZERO_OR_MORE,
            PatternWildcard.BRANCH_LABEL,  # LOOP2
            GenericInstr.BEQ,
            PatternWildcard.ANY_ZERO_OR_MORE,
            GenericInstr.BNE,
            PatternWildcard.ANY_ZERO_OR_MORE,
            PatternWildcard.BRANCH_LABEL,  # IF_EXIT2
            PatternWildcard.ANY_ZERO_OR_MORE,
            GenericInstr.JMP,
            PatternWildcard.BRANCH_LABEL,  # LOOP_EXIT2
            PatternWildcard.ANY_ZERO_OR_MORE,
            GenericInstr.BNE,
            PatternWildcard.ANY_ZERO_OR_MORE,
            PatternWildcard.BRANCH_LABEL,  # LOOP3
            GenericInstr.BEQ,
            PatternWildcard.ANY_ZERO_OR_MORE,
            GenericInstr.JMP,
            PatternWildcard.BRANCH_LABEL,  # LOOP_EXIT3
            PatternWildcard.BRANCH_LABEL,  # IF_EXIT3
        ]
    )


def test_epr_create_keep_info():
    DebugConnection.node_ids = {
        "Alice": 0,
        "Bob": 1,
    }

    epr_socket = EPRSocket("Bob")

    with DebugConnection("Alice", epr_sockets=[epr_socket]) as conn:
        eprs, infos = epr_socket.create_keep_with_info()
        epr, info = eprs[0], infos[0]

        _ = epr.measure(store_array=False)
        with info.generation_duration.if_ge(1337):
            _ = Qubit(conn)

        subroutine = conn._builder.subrt_pop_pending_subroutine()
        print(subroutine)

    inspector = ProtoSubroutineInspector(subroutine)
    assert inspector.match_pattern(
        [
            GenericInstr.CREATE_EPR,
            GenericInstr.WAIT_ALL,
            PatternWildcard.ANY_ZERO_OR_MORE,
            GenericInstr.LOAD,
            GenericInstr.BLT,
            GenericInstr.SET,
            GenericInstr.QALLOC,
            GenericInstr.INIT,
            PatternWildcard.BRANCH_LABEL,
        ]
    )


def test_epr_recv_keep_info():
    DebugConnection.node_ids = {
        "Alice": 0,
        "Bob": 1,
    }

    epr_socket = EPRSocket("Bob")

    with DebugConnection("Alice", epr_sockets=[epr_socket]) as conn:
        eprs, infos = epr_socket.recv_keep_with_info()
        epr, info = eprs[0], infos[0]

        _ = epr.measure(store_array=False)
        with info.generation_duration.if_ge(1337):
            _ = Qubit(conn)

        subroutine = conn._builder.subrt_pop_pending_subroutine()
        print(subroutine)

    inspector = ProtoSubroutineInspector(subroutine)
    assert inspector.match_pattern(
        [
            GenericInstr.RECV_EPR,
            GenericInstr.WAIT_ALL,
            PatternWildcard.ANY_ZERO_OR_MORE,  # Bell corrections
            GenericInstr.ROT_Z,  # Bell corrections
            PatternWildcard.ANY_ZERO_OR_MORE,  # Bell corrections
            GenericInstr.LOAD,
            GenericInstr.BLT,
            GenericInstr.SET,
            GenericInstr.QALLOC,
            GenericInstr.INIT,
            PatternWildcard.BRANCH_LABEL,
        ]
    )


def test_epr_recv_keep_info_no_phi_plus():
    DebugConnection.node_ids = {
        "Alice": 0,
        "Bob": 1,
    }

    epr_socket = EPRSocket("Bob")

    with DebugConnection("Alice", epr_sockets=[epr_socket]) as conn:
        eprs, infos = epr_socket.recv_keep_with_info(expect_phi_plus=False)
        epr, info = eprs[0], infos[0]

        _ = epr.measure(store_array=False)
        with info.generation_duration.if_ge(1337):
            _ = Qubit(conn)

        subroutine = conn._builder.subrt_pop_pending_subroutine()
        print(subroutine)

    inspector = ProtoSubroutineInspector(subroutine)
    assert inspector.match_pattern(
        [
            GenericInstr.RECV_EPR,
            GenericInstr.WAIT_ALL,
            PatternWildcard.ANY_ZERO_OR_MORE,
            GenericInstr.LOAD,
            GenericInstr.BLT,
            GenericInstr.SET,
            GenericInstr.QALLOC,
            GenericInstr.INIT,
            PatternWildcard.BRANCH_LABEL,
        ]
    )


def test_epr_context():
    DebugConnection.node_ids = {
        "Alice": 0,
        "Bob": 1,
    }

    epr_socket = EPRSocket("Bob")

    with DebugConnection("Alice", epr_sockets=[epr_socket]) as conn:

        with epr_socket.create_context(5) as (qubit, index):
            with index.if_eq(1337):
                qubit.H()
            _ = qubit.measure()

        subroutine = conn._builder.subrt_pop_pending_subroutine()
        print(subroutine)

    inspector = ProtoSubroutineInspector(subroutine)
    assert inspector.match_pattern(
        [
            GenericInstr.CREATE_EPR,
            PatternWildcard.ANY_ZERO_OR_MORE,
            PatternWildcard.BRANCH_LABEL,
            PatternWildcard.ANY_ZERO_OR_MORE,
            GenericInstr.WAIT_ALL,
            GenericInstr.BNE,
            GenericInstr.LOAD,
            GenericInstr.H,
            PatternWildcard.BRANCH_LABEL,
            GenericInstr.LOAD,
            GenericInstr.MEAS,
            PatternWildcard.ANY_ZERO_OR_MORE,
            PatternWildcard.BRANCH_LABEL,
        ]
    )


def test_epr_context_future_index():
    DebugConnection.node_ids = {
        "Alice": 0,
        "Bob": 1,
    }

    epr_socket = EPRSocket("Bob")

    with DebugConnection("Alice", epr_sockets=[epr_socket]) as conn:
        outcomes = conn.new_array(5)

        with epr_socket.create_context(5) as (qubit, index):
            outcome = outcomes.get_future_index(index)
            qubit.measure(outcome)

        subroutine = conn._builder.subrt_pop_pending_subroutine()
        print(subroutine)

    inspector = ProtoSubroutineInspector(subroutine)
    assert inspector.match_pattern(
        [
            GenericInstr.CREATE_EPR,
            PatternWildcard.ANY_ZERO_OR_MORE,
            PatternWildcard.BRANCH_LABEL,
            PatternWildcard.ANY_ZERO_OR_MORE,
            GenericInstr.WAIT_ALL,
            GenericInstr.LOAD,
            GenericInstr.MEAS,
            GenericInstr.QFREE,
            GenericInstr.STORE,
            GenericInstr.ADD,
            GenericInstr.JMP,
            PatternWildcard.BRANCH_LABEL,
        ]
    )


def test_create_epr_post():
    DebugConnection.node_ids = {
        "Alice": 0,
        "Bob": 1,
    }

    epr_socket = EPRSocket("Bob")

    with DebugConnection("Alice", epr_sockets=[epr_socket]) as conn:
        outcomes = conn.new_array(10)

        def post_create(_: DebugConnection, q: Qubit, index: RegFuture):
            q.H()
            index.add(1)
            outcome = outcomes.get_future_index(index)
            q.measure(outcome)

        epr_socket.create_keep(number=10, post_routine=post_create, sequential=True)

        subroutine = conn._builder.subrt_pop_pending_subroutine()
        print(subroutine)

    inspector = ProtoSubroutineInspector(subroutine)
    assert inspector.match_pattern(
        [
            GenericInstr.CREATE_EPR,
            PatternWildcard.ANY_ZERO_OR_MORE,
            PatternWildcard.BRANCH_LABEL,
            PatternWildcard.ANY_ZERO_OR_MORE,
            GenericInstr.WAIT_ALL,
            PatternWildcard.ANY_ZERO_OR_MORE,
            GenericInstr.LOAD,
            GenericInstr.H,
            GenericInstr.ADD,
            GenericInstr.LOAD,
            GenericInstr.MEAS,
            GenericInstr.QFREE,
            GenericInstr.STORE,
            GenericInstr.ADD,
            GenericInstr.JMP,
            PatternWildcard.BRANCH_LABEL,
        ]
    )


def test_recv_epr_post():
    DebugConnection.node_ids = {
        "Alice": 0,
        "Bob": 1,
    }

    epr_socket = EPRSocket("Bob")

    with DebugConnection("Alice", epr_sockets=[epr_socket]) as conn:
        outcomes = conn.new_array(10)

        def post_create(_: DebugConnection, q: Qubit, index: RegFuture):
            q.H()
            index.add(1)
            outcome = outcomes.get_future_index(index)
            q.measure(outcome)

        epr_socket.recv_keep(number=10, post_routine=post_create, sequential=True)

        subroutine = conn._builder.subrt_pop_pending_subroutine()
        print(subroutine)

    inspector = ProtoSubroutineInspector(subroutine)
    assert inspector.match_pattern(
        [
            GenericInstr.RECV_EPR,
            PatternWildcard.ANY_ZERO_OR_MORE,
            PatternWildcard.BRANCH_LABEL,
            PatternWildcard.ANY_ZERO_OR_MORE,
            GenericInstr.WAIT_ALL,
            PatternWildcard.ANY_ZERO_OR_MORE,  # Bell corrections
            GenericInstr.ROT_Z,  # Bell corrections
            PatternWildcard.ANY_ZERO_OR_MORE,  # Bell corrections
            GenericInstr.LOAD,
            GenericInstr.H,
            GenericInstr.ADD,
            GenericInstr.LOAD,
            GenericInstr.MEAS,
            GenericInstr.QFREE,
            GenericInstr.STORE,
            GenericInstr.ADD,
            GenericInstr.JMP,
            PatternWildcard.BRANCH_LABEL,
        ]
    )


def test_try():
    with DebugConnection("Alice") as conn:

        with conn.try_until_success(max_tries=1):
            q = Qubit(conn)
            q.measure()

        subroutine = conn.builder.subrt_pop_pending_subroutine()
        print(subroutine)

    inspector = ProtoSubroutineInspector(subroutine)
    assert inspector.match_pattern(
        [
            GenericInstr.QALLOC,
            GenericInstr.INIT,
            PatternWildcard.ANY_ZERO_OR_MORE,
            GenericInstr.MEAS,
            GenericInstr.QFREE,
            GenericInstr.STORE,
        ]
    )


def test_loop_until():
    with DebugConnection("Alice") as conn:

        with conn.loop_until(max_iterations=10) as loop:
            q = Qubit(conn)
            m = q.measure()
            constraint = ValueAtMostConstraint(m, 42)
            loop.set_exit_condition(constraint)

        subroutine = conn.builder.subrt_pop_pending_subroutine()
        print(subroutine)

    inspector = ProtoSubroutineInspector(subroutine)
    assert inspector.match_pattern(
        [
            PatternWildcard.BRANCH_LABEL,
            GenericInstr.BEQ,
            GenericInstr.SET,
            GenericInstr.QALLOC,
            GenericInstr.INIT,
            PatternWildcard.ANY_ZERO_OR_MORE,
            GenericInstr.MEAS,
            GenericInstr.QFREE,
            GenericInstr.STORE,
            GenericInstr.LOAD,
            GenericInstr.BLT,
            GenericInstr.ADD,
            GenericInstr.JMP,
            PatternWildcard.BRANCH_LABEL,
        ]
    )


def test_create_epr_min_fidelity_all():
    DebugConnection.node_ids = {
        "Alice": 0,
        "Bob": 1,
    }

    epr_socket = EPRSocket("Bob")

    with DebugConnection("Alice", epr_sockets=[epr_socket]) as conn:
        epr_socket.create_keep(number=2, min_fidelity_all_at_end=80, max_tries=100)

        subroutine = conn._builder.subrt_pop_pending_subroutine()
        print(subroutine)

    inspector = ProtoSubroutineInspector(subroutine)
    assert inspector.match_pattern(
        [
            GenericInstr.BEQ,
            GenericInstr.SET,
            PatternWildcard.BRANCH_LABEL,
            GenericInstr.BEQ,
            GenericInstr.UNDEF,
            GenericInstr.ADD,
            GenericInstr.JMP,
            PatternWildcard.BRANCH_LABEL,
            GenericInstr.CREATE_EPR,
            GenericInstr.WAIT_ALL,
            PatternWildcard.ANY_ZERO_OR_MORE,
            GenericInstr.LOAD,
            GenericInstr.BLT,
            PatternWildcard.ANY_ZERO_OR_MORE,
            GenericInstr.UNDEF,
            PatternWildcard.ANY_ZERO_OR_MORE,
            GenericInstr.JMP,
            PatternWildcard.BRANCH_LABEL,
        ]
    )


def test_recv_epr_min_fidelity_all():
    DebugConnection.node_ids = {
        "Alice": 0,
        "Bob": 1,
    }

    epr_socket = EPRSocket("Bob")

    with DebugConnection("Alice", epr_sockets=[epr_socket]) as conn:
        epr_socket.recv_keep(number=2, min_fidelity_all_at_end=80, max_tries=100)

        subroutine = conn._builder.subrt_pop_pending_subroutine()
        print(subroutine)

    inspector = ProtoSubroutineInspector(subroutine)
    assert inspector.match_pattern(
        [
            GenericInstr.BEQ,
            GenericInstr.SET,
            PatternWildcard.BRANCH_LABEL,
            GenericInstr.BEQ,
            GenericInstr.UNDEF,
            GenericInstr.ADD,
            GenericInstr.JMP,
            PatternWildcard.BRANCH_LABEL,
            GenericInstr.RECV_EPR,
            GenericInstr.WAIT_ALL,
            PatternWildcard.ANY_ZERO_OR_MORE,  # Bell corrections
            GenericInstr.ROT_Z,  # Bell corrections
            PatternWildcard.ANY_ZERO_OR_MORE,  # Bell corrections
            GenericInstr.LOAD,
            GenericInstr.BLT,
            PatternWildcard.ANY_ZERO_OR_MORE,
            GenericInstr.UNDEF,
            PatternWildcard.ANY_ZERO_OR_MORE,
            GenericInstr.JMP,
            PatternWildcard.BRANCH_LABEL,
        ]
    )


def test_create_multiple_eprs():
    DebugConnection.node_ids = {
        "Alice": 0,
        "Bob": 1,
    }

    epr_socket = EPRSocket("Bob")

    with DebugConnection("Alice", epr_sockets=[epr_socket]) as conn:
        epr0, epr1 = epr_socket.create(2)

        epr0.X()
        epr1.Y()

        _ = epr0.measure(store_array=False)
        _ = epr1.measure(store_array=False)

        subroutine = conn.builder.subrt_pop_pending_subroutine()
        print(subroutine)

    # Check if the qubits have the correct ID
    epr0_id = 0
    epr1_id = 1

    for i, cmd in enumerate(subroutine.commands):
        if not isinstance(cmd, ICmd):
            continue
        if cmd.instruction == GenericInstr.X:
            prev_set_instr = subroutine.commands[i - 1]
            assert prev_set_instr.operands[1] == epr0_id
        elif cmd.instruction == GenericInstr.Y:
            prev_set_instr = subroutine.commands[i - 1]
            assert prev_set_instr.operands[1] == epr1_id


def test_create_2_eprs_NV():
    DebugConnection.node_ids = {
        "Alice": 0,
        "Bob": 1,
    }

    epr_socket = EPRSocket("Bob")

    with DebugConnection(
        "Alice",
        epr_sockets=[epr_socket],
        hardware_config=NVHardwareConfig(2),
    ) as conn:
        epr0, epr1 = epr_socket.create_keep(2)
        print(f"epr0 ID: {epr0.qubit_id}")
        print(f"epr1 ID: {epr1.qubit_id}")

        epr0.X()
        epr1.Y()

        _ = epr0.measure(store_array=False)
        _ = epr1.measure(store_array=False)

        subroutine = conn.builder.subrt_pop_pending_subroutine()
        print(subroutine)

    # Check if the qubits have the correct ID
    epr0_id = 1
    epr1_id = 0

    for i, cmd in enumerate(subroutine.commands):
        if not isinstance(cmd, ICmd):
            continue
        if cmd.instruction == GenericInstr.X:
            prev_set_instr = subroutine.commands[i - 1]
            assert prev_set_instr.operands[1] == epr0_id
        elif cmd.instruction == GenericInstr.Y:
            prev_set_instr = subroutine.commands[i - 1]
            assert prev_set_instr.operands[1] == epr1_id


def test_create_3_eprs_NV():
    DebugConnection.node_ids = {
        "Alice": 0,
        "Bob": 1,
    }

    epr_socket = EPRSocket("Bob")

    with DebugConnection(
        "Alice",
        epr_sockets=[epr_socket],
        hardware_config=NVHardwareConfig(3),
    ) as conn:
        epr0, epr1, epr2 = epr_socket.create_keep(3)
        print(f"epr0 ID: {epr0.qubit_id}")
        print(f"epr1 ID: {epr1.qubit_id}")
        print(f"epr2 ID: {epr2.qubit_id}")

        epr0.X()
        epr1.Y()
        epr2.Z()

        _ = epr0.measure(store_array=False)
        _ = epr1.measure(store_array=False)
        _ = epr2.measure(store_array=False)

        subroutine = conn.builder.subrt_pop_pending_subroutine()
        print(subroutine)

    # Check if the qubits have the correct ID
    epr0_id = 2
    epr1_id = 1
    epr2_id = 0

    for i, cmd in enumerate(subroutine.commands):
        if not isinstance(cmd, ICmd):
            continue
        if cmd.instruction == GenericInstr.X:
            prev_set_instr = subroutine.commands[i - 1]
            assert prev_set_instr.operands[1] == epr0_id
        elif cmd.instruction == GenericInstr.Y:
            prev_set_instr = subroutine.commands[i - 1]
            assert prev_set_instr.operands[1] == epr1_id
        elif cmd.instruction == GenericInstr.Z:
            prev_set_instr = subroutine.commands[i - 1]
            assert prev_set_instr.operands[1] == epr2_id


def test_bqc_receiver_NV():
    DebugConnection.node_ids = {
        "Alice": 0,
        "Bob": 1,
    }

    epr_socket = EPRSocket("Bob")

    with DebugConnection(
        "Alice",
        epr_sockets=[epr_socket],
        hardware_config=NVHardwareConfig(2),
        compiler=NVSubroutineTranspiler,
    ) as conn:
        carbon, electron = epr_socket.recv_keep(2)

        electron.cphase(carbon)

        delta1 = math.pi
        electron.rot_Z(angle=delta1)
        electron.H()
        _ = electron.measure(store_array=False)

        delta2 = math.pi
        carbon.rot_Z(angle=delta2)
        carbon.H()
        _ = carbon.measure()

        protosubroutine = conn.builder.subrt_pop_pending_subroutine()
        print(protosubroutine)
        compiled_subroutine = conn.builder.subrt_compile_subroutine(protosubroutine)
        print(compiled_subroutine)


def test_bqc_receiver_NV_min_fidelity():
    DebugConnection.node_ids = {
        "Alice": 0,
        "Bob": 1,
    }

    epr_socket = EPRSocket("Bob")

    with DebugConnection(
        "Alice",
        epr_sockets=[epr_socket],
        hardware_config=NVHardwareConfig(2),
        compiler=NVSubroutineTranspiler,
    ) as conn:
        carbon, electron = epr_socket.recv_keep(
            2, min_fidelity_all_at_end=80, max_tries=25
        )

        electron.cphase(carbon)

        delta1 = math.pi
        electron.rot_Z(angle=delta1)
        electron.H()
        _ = electron.measure(store_array=False)

        delta2 = math.pi
        carbon.rot_Z(angle=delta2)
        carbon.H()
        _ = carbon.measure()

        protosubroutine = conn.builder.subrt_pop_pending_subroutine()
        print(protosubroutine)
        compiled_subroutine = conn.builder.subrt_compile_subroutine(protosubroutine)
        print(compiled_subroutine)


def test_measure_Z():
    with DebugConnection("Alice") as conn:
        q = Qubit(conn)
        q.measure(basis=QubitMeasureBasis.Z)

        presubroutine = conn.builder.subrt_pop_pending_subroutine()
        print(presubroutine)
        inspector = ProtoSubroutineInspector(presubroutine)
        assert inspector.contains_instr(GenericInstr.MEAS)

        compiled_subroutine = conn.builder.subrt_compile_subroutine(presubroutine)
        print(compiled_subroutine)


def test_measure_basis():
    with DebugConnection("Alice") as conn:
        q = Qubit(conn)
        q.measure(basis=QubitMeasureBasis.X)

        presubroutine = conn.builder.subrt_pop_pending_subroutine()
        print(presubroutine)
        inspector = ProtoSubroutineInspector(presubroutine)
        assert inspector.contains_instr(GenericInstr.MEAS_BASIS)

        [meas_basis] = inspector.find_instr(GenericInstr.MEAS_BASIS)
        # check if rotations are correct
        assert meas_basis.operands[2:] == [0, 24, 0, 4]  # skip register operands

        compiled_subroutine = conn.builder.subrt_compile_subroutine(presubroutine)
        print(compiled_subroutine)


def test_measure_basis_rotation():
    with DebugConnection("Alice") as conn:
        q = Qubit(conn)
        q.measure(basis_rotations=(3, 4, 5))

        presubroutine = conn.builder.subrt_pop_pending_subroutine()
        print(presubroutine)
        inspector = ProtoSubroutineInspector(presubroutine)
        assert inspector.contains_instr(GenericInstr.MEAS_BASIS)

        [meas_basis] = inspector.find_instr(GenericInstr.MEAS_BASIS)
        # check if rotations are correct
        assert meas_basis.operands[2:] == [3, 4, 5, 4]  # skip register operands

        compiled_subroutine = conn.builder.subrt_compile_subroutine(presubroutine)
        print(compiled_subroutine)


def test_create_keep_no_corrections():
    DebugConnection.node_ids = {
        "Alice": 0,
        "Bob": 1,
    }

    epr_socket = EPRSocket("Bob")

    with DebugConnection("Alice", epr_sockets=[epr_socket]) as conn:
        epr_socket.create_keep(number=1)

        subroutine = conn._builder.subrt_pop_pending_subroutine()
        print(subroutine)

    inspector = ProtoSubroutineInspector(subroutine)
    assert inspector.match_pattern(
        [
            GenericInstr.CREATE_EPR,
            GenericInstr.WAIT_ALL,
            GenericInstr.RET_ARR,
        ]
    )


def test_recv_keep_no_corrections():
    DebugConnection.node_ids = {
        "Alice": 0,
        "Bob": 1,
    }

    epr_socket = EPRSocket("Bob")

    with DebugConnection("Alice", epr_sockets=[epr_socket]) as conn:
        epr_socket.recv_keep(number=1, expect_phi_plus=False)

        subroutine = conn._builder.subrt_pop_pending_subroutine()
        print(subroutine)

    inspector = ProtoSubroutineInspector(subroutine)
    assert inspector.match_pattern(
        [
            GenericInstr.RECV_EPR,
            GenericInstr.WAIT_ALL,
            GenericInstr.RET_ARR,
        ]
    )


def test_recv_keep_corrections():
    DebugConnection.node_ids = {
        "Alice": 0,
        "Bob": 1,
    }

    epr_socket = EPRSocket("Bob")

    with DebugConnection("Alice", epr_sockets=[epr_socket]) as conn:
        epr_socket.recv_keep(number=1, expect_phi_plus=True)

        subroutine = conn._builder.subrt_pop_pending_subroutine()
        print(subroutine)

    inspector = ProtoSubroutineInspector(subroutine)
    assert inspector.match_pattern(
        [
            GenericInstr.RECV_EPR,
            GenericInstr.WAIT_ALL,
            GenericInstr.SET,
            PatternWildcard.ANY_ZERO_OR_MORE,
            GenericInstr.BNE,
            PatternWildcard.ANY_ZERO_OR_MORE,
            GenericInstr.RET_ARR,
        ]
    )


def test_create_measure_no_corrections():
    DebugConnection.node_ids = {
        "Alice": 0,
        "Bob": 1,
    }

    epr_socket = EPRSocket("Bob")

    with DebugConnection("Alice", epr_sockets=[epr_socket]) as conn:
        epr_socket.create_measure(number=1)

        subroutine = conn._builder.subrt_pop_pending_subroutine()
        print(subroutine)

    inspector = ProtoSubroutineInspector(subroutine)
    assert inspector.match_pattern(
        [
            GenericInstr.CREATE_EPR,
            GenericInstr.WAIT_ALL,
            GenericInstr.RET_ARR,
        ]
    )


def test_recv_measure_no_corrections():
    DebugConnection.node_ids = {
        "Alice": 0,
        "Bob": 1,
    }

    epr_socket = EPRSocket("Bob")

    with DebugConnection("Alice", epr_sockets=[epr_socket]) as conn:
        epr_socket.recv_measure(number=1, expect_phi_plus=False)

        subroutine = conn._builder.subrt_pop_pending_subroutine()
        print(subroutine)

    inspector = ProtoSubroutineInspector(subroutine)
    assert inspector.match_pattern(
        [
            GenericInstr.RECV_EPR,
            GenericInstr.WAIT_ALL,
            GenericInstr.RET_ARR,
        ]
    )


def test_recv_measure_corrections():
    DebugConnection.node_ids = {
        "Alice": 0,
        "Bob": 1,
    }

    epr_socket = EPRSocket("Bob")

    with DebugConnection("Alice", epr_sockets=[epr_socket]) as conn:
        epr_socket.recv_measure(number=1, expect_phi_plus=True)

        subroutine = conn._builder.subrt_pop_pending_subroutine()
        print(subroutine)

    inspector = ProtoSubroutineInspector(subroutine)
    assert inspector.match_pattern(
        [
            GenericInstr.RECV_EPR,
            GenericInstr.WAIT_ALL,
            GenericInstr.RET_ARR,
        ]
    )


def test_create_rsp_no_corrections():
    DebugConnection.node_ids = {
        "Alice": 0,
        "Bob": 1,
    }

    epr_socket = EPRSocket("Bob")

    with DebugConnection("Alice", epr_sockets=[epr_socket]) as conn:
        epr_socket.create_rsp(number=1)

        subroutine = conn._builder.subrt_pop_pending_subroutine()
        print(subroutine)

    inspector = ProtoSubroutineInspector(subroutine)
    assert inspector.match_pattern(
        [
            GenericInstr.CREATE_EPR,
            GenericInstr.WAIT_ALL,
            GenericInstr.RET_ARR,
        ]
    )


def test_recv_rsp_no_corrections():
    DebugConnection.node_ids = {
        "Alice": 0,
        "Bob": 1,
    }

    epr_socket = EPRSocket("Bob")

    with DebugConnection("Alice", epr_sockets=[epr_socket]) as conn:
        epr_socket.recv_rsp(number=1, expect_phi_plus=False)

        subroutine = conn._builder.subrt_pop_pending_subroutine()
        print(subroutine)

    inspector = ProtoSubroutineInspector(subroutine)
    assert inspector.match_pattern(
        [
            GenericInstr.RECV_EPR,
            GenericInstr.WAIT_ALL,
            GenericInstr.RET_ARR,
        ]
    )


def test_recv_rsp_corrections():
    DebugConnection.node_ids = {
        "Alice": 0,
        "Bob": 1,
    }

    epr_socket = EPRSocket("Bob")

    with DebugConnection("Alice", epr_sockets=[epr_socket]) as conn:
        epr_socket.recv_rsp(number=1, expect_phi_plus=True)

        subroutine = conn._builder.subrt_pop_pending_subroutine()
        print(subroutine)

    inspector = ProtoSubroutineInspector(subroutine)
    assert inspector.match_pattern(
        [
            GenericInstr.RECV_EPR,
            GenericInstr.WAIT_ALL,
            GenericInstr.SET,
            PatternWildcard.ANY_ZERO_OR_MORE,
            GenericInstr.BNE,
            PatternWildcard.ANY_ZERO_OR_MORE,
            GenericInstr.RET_ARR,
        ]
    )


if __name__ == "__main__":
    # set_log_level("DEBUG")
    test_simple()
    test_create_epr()
    test_branching()
    test_loop_context()
    test_looping()
    test_futures()
    test_nested()
    test_epr_create_keep_info()
    test_epr_recv_keep_info()
    test_epr_recv_keep_info_no_phi_plus()
    test_epr_context()
    test_epr_context_future_index()
    test_create_epr_post()
    test_recv_epr_post()
    test_try()
    test_loop_until()
    test_create_epr_min_fidelity_all()
    test_recv_epr_min_fidelity_all()
    test_create_multiple_eprs()
    test_create_2_eprs_NV()
    test_create_3_eprs_NV()
    test_bqc_receiver_NV()
    test_bqc_receiver_NV_min_fidelity()
    test_measure_Z()
    test_measure_basis()
    test_measure_basis_rotation()

    test_create_keep_no_corrections()
    test_recv_keep_no_corrections()
    test_recv_keep_corrections()

    test_create_measure_no_corrections()
    test_recv_measure_no_corrections()
    test_recv_measure_corrections()

    test_create_rsp_no_corrections()
    test_recv_rsp_no_corrections()
    test_recv_rsp_corrections()



import inspect
import logging
import os
import runpy
import subprocess
import sys
from collections import namedtuple

from netqasm.logging.glob import set_log_level
from netqasm.runtime.settings import Simulator, get_simulator

# Used to specify to skip some apps in some cases
SkipIf = namedtuple("SkipIf", ["skip", "reason"])


skip_ifs = {
    "anonymous_transmission": SkipIf(
        get_simulator() == Simulator.SIMULAQRON,
        reason="Current broadcast channel in simulaqron cannot handle socket IDs",
    ),
    "link_layer_md": SkipIf(
        get_simulator() == Simulator.NETSQUID
        or get_simulator() == Simulator.SIMULAQRON,
        reason="Current simulators do not handle MD requests well.",
    ),
}


def _has_first_argument(function, argument):
    """Checks if a function takes a named argument as the first argument"""
    argnames = inspect.getfullargspec(function).args
    return argnames[0] == "no_output"


def main(external):
    set_log_level(logging.WARNING)
    path_to_here = os.path.dirname(os.path.abspath(__file__))

    if external:
        apps_path = os.path.join(path_to_here, "apps")
        apps = os.listdir(apps_path)
        for app in apps:
            app_path = os.path.join(apps_path, app)
            if not os.path.isdir(app_path) or app == "__pycache__":
                continue
            skip_if = skip_ifs.get(app)
            if skip_if is not None and skip_if.skip:
                print(f"Skipping example app {app_path} since {skip_if.reason}")
                continue
            print(f"Running example app {app_path} using simulator {get_simulator()}")
            result = subprocess.run(
                ["netqasm", "simulate", "--app-dir", app_path],
                stdout=subprocess.DEVNULL,
            )
            if result.returncode != 0:
                raise RuntimeError(f"Example {app} failed!")
    else:
        for root, _folders, files in os.walk(path_to_here):
            for filename in files:
                if filename.startswith("example") and filename.endswith(".py"):
                    filepath = os.path.join(root, filename)
                    members = runpy.run_path(filepath)
                    if "main" in members:
                        print(f"Running example {filepath}")
                        result = subprocess.run(
                            ["python3", filepath],
                            stdout=subprocess.DEVNULL,
                        )
                        if result.returncode != 0:
                            raise RuntimeError(f"Example {filepath} failed!")
                    else:
                        print(f"The example {filepath} does not have a main function")
    print("All examples work!")


if __name__ == "__main__":
    external = "--external" in sys.argv
    main(external=external)



from collections import defaultdict
from itertools import count
from typing import Dict, List, Optional, Set, Tuple, Union

from netqasm.lang.encoding import REG_INDEX_BITS, RegisterName
from netqasm.lang.instr import Flavour, VanillaFlavour
from netqasm.lang.instr.base import NetQASMInstruction
from netqasm.lang.ir import (
    BranchLabel,
    GenericInstr,
    ICmd,
    ProtoSubroutine,
    string_to_instruction,
)
from netqasm.lang.operand import (
    Address,
    ArrayEntry,
    ArraySlice,
    Label,
    Register,
    Template,
)
from netqasm.lang.subroutine import Subroutine
from netqasm.lang.symbols import Symbols
from netqasm.util.error import NetQASMInstrError, NetQASMSyntaxError
from netqasm.util.string import group_by_word, is_number, is_variable_name

T_Cmd = Union[ICmd, BranchLabel]
T_ParsedValue = Union[int, Register, Label, Template, str]


def parse_text_protosubroutine(text: str) -> ProtoSubroutine:
    """
    Convert a text representation of a subroutine into a ProtoSubroutine object.
    """
    preamble_lines, body_lines_with_macros = _split_preamble_body(text)
    preamble_data = _parse_preamble(preamble_lines)
    body_lines = _apply_macros(
        body_lines_with_macros, preamble_data[Symbols.PREAMBLE_DEFINE]
    )
    return _create_subroutine(preamble_data, body_lines)


def parse_text_subroutine(
    subroutine: str,
    assign_branch_labels=True,
    make_args_operands=True,
    replace_constants=True,
    flavour: Flavour = None,
) -> Subroutine:
    """
    Convert a text representation of a subroutine into a Subroutine object.

    Internally, first a `ProtoSubroutine` object is created, consisting of `ICmd`s.
    This is then converted into a `Subroutine` using `assemble_subroutine`.
    """
    pre_subroutine = parse_text_protosubroutine(subroutine)
    assembled_subroutine = assemble_subroutine(
        pre_subroutine=pre_subroutine,
        assign_branch_labels=assign_branch_labels,
        make_args_operands=make_args_operands,
        replace_constants=replace_constants,
        flavour=flavour,
    )
    return assembled_subroutine


def assemble_subroutine(
    pre_subroutine: ProtoSubroutine,
    assign_branch_labels=True,
    make_args_operands=True,
    replace_constants=True,
    flavour: Flavour = None,
) -> Subroutine:
    """
    Convert a `ProtoSubroutine` into a `Subroutine`, given a Flavour (default: vanilla).
    """
    if make_args_operands:
        _make_args_operands(pre_subroutine)
    if replace_constants:
        pre_subroutine.commands = _replace_constants(pre_subroutine.commands)
    if assign_branch_labels:
        _assign_branch_labels(pre_subroutine)

    if flavour is None:
        flavour = VanillaFlavour()
    subroutine = _build_subroutine(pre_subroutine, flavour)

    return subroutine


def _build_subroutine(pre_subroutine: ProtoSubroutine, flavour: Flavour) -> Subroutine:
    instructions: List[NetQASMInstruction] = []

    for command in pre_subroutine.commands:
        assert isinstance(command, ICmd)

        instr = flavour.get_instr_by_name(command.instruction.name.lower())
        new_command = instr.from_operands(command.operands)
        new_command.lineno = command.lineno
        instructions.append(new_command)

    return Subroutine(
        instructions=instructions,
        arguments=pre_subroutine.arguments,
        netqasm_version=pre_subroutine.netqasm_version,
        app_id=pre_subroutine.app_id,
    )


def _create_subroutine(
    preamble_data: Dict[str, List[List[str]]], body_lines: List[str]
) -> ProtoSubroutine:
    commands: List[Union[ICmd, BranchLabel]] = []
    for line in body_lines:
        if line.endswith(Symbols.BRANCH_END):
            # A command defining a branch label should end with BRANCH_END
            branch_label = line.rstrip(Symbols.BRANCH_END)
            if not is_variable_name(branch_label):
                raise NetQASMSyntaxError(
                    f"The branch label {branch_label} is not a valid label"
                )
            commands.append(BranchLabel(branch_label))
        else:
            words: List[str] = group_by_word(line, brackets=Symbols.ARGS_BRACKETS)
            instr_name, args = _split_instr_and_args(words[0])
            instr = string_to_instruction(instr_name)
            args = _parse_args(args)
            if instr in _ALLOW_LABEL_INSTRUCTIONS:
                operands = _parse_operands(words[1:], allow_label=True)
            else:
                operands = _parse_operands(words[1:], allow_label=False)
            command = ICmd(
                instruction=instr,
                args=args,
                operands=operands,
            )
            commands.append(command)

    if Symbols.PREAMBLE_NETQASM in preamble_data:
        netqasm_version = _parse_netqasm_version(
            preamble_data[Symbols.PREAMBLE_NETQASM][0][0]
        )
    else:
        netqasm_version = None

    app_id: Optional[int]
    if Symbols.PREAMBLE_APPID in preamble_data:
        app_id = int(preamble_data[Symbols.PREAMBLE_APPID][0][0])
    else:
        app_id = None

    return ProtoSubroutine(
        commands=commands,
        netqasm_version=netqasm_version,
        app_id=app_id,
    )


def _parse_netqasm_version(netqasm_version):
    try:
        major, minor = netqasm_version.strip().split(".")
        return int(major), int(minor)
    except Exception as err:
        raise ValueError(
            f"Could not parse netqasm version {netqasm_version} since: {err}"
        )


def _split_instr_and_args(word):
    instr, args = _split_of_bracket(word, brackets=Symbols.ARGS_BRACKETS)
    return instr, args


def _parse_args(args):
    if args == "":
        return []
    else:
        return [
            _parse_constant(arg.strip())
            for arg in args.strip(Symbols.ARGS_BRACKETS).split(Symbols.ARGS_DELIM)
        ]


def _parse_constant(constant: str) -> int:
    if not is_number(constant):
        raise NetQASMSyntaxError(f"Expected constant, got {constant}")
    return int(constant)


def _parse_operands(words: List[str], allow_label=True):
    operands = []
    for word in words:
        operand = _parse_operand(word.strip(), allow_label=allow_label)
        operands.append(operand)

    return operands


def _parse_operand(word: str, allow_label=True):
    if word.startswith(Symbols.ADDRESS_START):
        return parse_address(word)
    else:
        return _parse_value(word, allow_label=allow_label, allow_template=True)


def _parse_value(
    value: str,
    allow_label: bool = False,
    allow_template: bool = False,
    allow_keywords: Optional[List[str]] = None,
) -> T_ParsedValue:
    # Try to parse a constant
    try:
        return _parse_constant(value)
    except NetQASMSyntaxError:
        pass

    # Try to parse a register
    try:
        return parse_register(value)
    except NetQASMSyntaxError:
        pass

    if allow_keywords:
        for keyword in allow_keywords:
            if value == keyword:
                return value

    if allow_label:
        # Parse a label
        try:
            return _parse_label(value)
        except NetQASMSyntaxError:
            pass

    if allow_template:
        # Parse a template
        try:
            return _parse_template(value)
        except NetQASMSyntaxError:
            pass

    raise NetQASMSyntaxError(f"{value} is not a valid value in this case")


def _is_byte(value):
    if not value.startswith("0x"):
        return False
    if not len(value) == 4:
        return False
    return is_number(value[2:])


def _parse_label(label: str) -> Label:
    if not is_variable_name(label):
        raise NetQASMSyntaxError(f"Expected a label, got {label}")
    return Label(label)


def _parse_template(template: str) -> Template:
    assert template.startswith("{")
    assert template.endswith("}")
    template = template.strip(Symbols.TEMPLATE_BRACKETS).strip()
    return Template(template)


_REGISTER_NAMES = {reg.name: reg for reg in RegisterName}


def parse_register(register: str) -> Register:
    try:
        register_name = _REGISTER_NAMES[register[0]]
    except KeyError:
        raise NetQASMSyntaxError(f"{register[0]} is not a valid register name")
    value = _parse_constant(register[1:])
    return Register(register_name, value)


def parse_address(address: str) -> Union[Address, ArraySlice, ArrayEntry]:
    base_address, index_str = _split_of_bracket(address, Symbols.INDEX_BRACKETS)
    base_address_int_or_str: Union[int, str] = _parse_base_address(base_address)
    index = _parse_index(index_str)
    address_parsed = Address(base_address_int_or_str)
    if index is None:
        return address_parsed
    elif isinstance(index, tuple):
        if (not isinstance(index[0], Register) and not isinstance(index[0], int)) or (
            not isinstance(index[1], Register) and not isinstance(index[1], int)
        ):
            raise TypeError(
                f"indices {index[0]} and {index[1]} should be ints or Registers,"
                f" not {type(index[0])} and {type(index[1])}"
            )
        return ArraySlice(address_parsed, start=index[0], stop=index[1])
    elif isinstance(index, int) or isinstance(index, Register):
        return ArrayEntry(address_parsed, index)
    else:
        raise TypeError(f"Index cannot have type {type(index)}")


def _parse_base_address(base_address: str) -> Union[int, str]:
    if not base_address.startswith(Symbols.ADDRESS_START):
        raise NetQASMSyntaxError(f"Expected address, got {base_address}")
    keywords = ["input", "output"]
    value = _parse_value(
        base_address.lstrip(Symbols.ADDRESS_START), allow_keywords=keywords
    )
    if not isinstance(value, int):
        if not isinstance(value, str) and value in keywords:
            raise TypeError(f"Address should be an int, not a {type(value)}")
    assert isinstance(value, int) or isinstance(value, str)
    return value


def _parse_index(
    index: str,
) -> Optional[Union[T_ParsedValue, Tuple[T_ParsedValue, T_ParsedValue]]]:
    if index == "":
        return None
    index = index.strip(Symbols.INDEX_BRACKETS).strip()
    if Symbols.SLICE_DELIM in index:
        start, stop = index.split(Symbols.SLICE_DELIM)
        return _parse_value(start.strip()), _parse_value(stop.strip())
    else:
        return _parse_value(index)


def _split_of_bracket(word: str, brackets: str) -> Tuple[str, str]:
    if len(brackets) != 2:
        raise ValueError(f"`brackets` cannot be {brackets}")
    start_bracket: str = brackets[0]
    end_bracket: str = brackets[1]
    start = word.find(start_bracket)
    if start == -1:
        return word, ""
    if word[-1] != end_bracket:
        raise NetQASMSyntaxError(f"No end bracket in {word}, expected '{end_bracket}'")
    address = word[:start]
    content = word[start:]
    return address, content


def _split_preamble_body(subroutine_text: str) -> Tuple[List[str], List[str]]:
    """Splits the preamble from the body of the subroutine"""
    is_preamble = True
    preamble_lines = []
    body_lines = []
    for line in subroutine_text.split("\n"):
        # Remove surrounding whitespace and comments
        line = line.strip()
        line = _remove_comments_from_line(line)
        if line == "":  # Ignore empty lines
            continue
        if line.startswith(Symbols.PREAMBLE_START):
            if not is_preamble:  # Should not go out of preamble and in again
                raise NetQASMSyntaxError(
                    "Cannot have a preamble line after instructions"
                )
            line = line.lstrip(Symbols.PREAMBLE_START).strip()
            if line == Symbols.PREAMBLE_START:  # Ignore lines with only a '#' character
                continue
            preamble_lines.append(line)
        else:
            is_preamble = (
                False  # From now on the lines should not be part of the preamble
            )
            body_lines.append(line)
    return preamble_lines, body_lines


def _apply_macros(body_lines, macros) -> List[str]:
    """Applies macros to the body lines"""
    if len(body_lines) == 0:
        return []
    body = "\n".join(body_lines)
    for macro_key, macro_value in macros:
        macro_value = macro_value.strip(Symbols.PREAMBLE_DEFINE_BRACKETS)
        body = body.replace(f"{Symbols.MACRO_START}{macro_key}", macro_value)
    return list(body.split("\n"))


def _remove_comments_from_line(line):
    """Removes comments from a line"""
    return line.split(Symbols.COMMENT_START)[0]


def _parse_preamble(preamble_lines: List[str]) -> Dict[str, List[List[str]]]:
    """Parses the preamble lines"""

    preamble_instructions = defaultdict(list)
    for line in preamble_lines:
        try:
            instr, *operands = group_by_word(
                line, brackets=Symbols.PREAMBLE_DEFINE_BRACKETS
            )
        except ValueError as err:
            raise NetQASMSyntaxError(
                f"Could not parse preamble instruction, since: {err}"
            )
        preamble_instructions[instr].append(operands)
    _assert_valid_preamble_instructions(preamble_instructions)
    return preamble_instructions


def _assert_valid_preamble_instructions(
    preamble_instructions, is_template: bool = False
):
    if is_template:
        preamble_assertions = {
            Symbols.PREAMBLE_DEFINE: _assert_valid_preamble_instr_define,
        }
    else:
        preamble_assertions = {
            Symbols.PREAMBLE_NETQASM: _assert_valid_preamble_instr_netqasm,
            Symbols.PREAMBLE_APPID: _assert_valid_preamble_instr_appid,
            Symbols.PREAMBLE_DEFINE: _assert_valid_preamble_instr_define,
        }
    for instr, list_of_operands in preamble_instructions.items():
        preamble_assertion = preamble_assertions.get(instr)
        if preamble_assertion is None:
            raise NetQASMInstrError(
                f"The instruction {instr} is not a valid preamble instruction"
            )
        preamble_assertion(list_of_operands)


def _assert_valid_preamble_instr_netqasm(list_of_operands):
    _assert_single_preamble_instr(list_of_operands, Symbols.PREAMBLE_NETQASM)
    _assert_single_preamble_arg(list_of_operands, Symbols.PREAMBLE_NETQASM)


def _assert_valid_preamble_instr_appid(list_of_operands):
    _assert_single_preamble_instr(list_of_operands, Symbols.PREAMBLE_APPID)
    _assert_single_preamble_arg(list_of_operands, Symbols.PREAMBLE_APPID)


def _assert_valid_preamble_instr_define(list_of_operands):
    macro_keys = []
    for operands in list_of_operands:
        if len(operands) != 2:
            raise NetQASMSyntaxError(
                f"Preamble instruction {Symbols.PREAMBLE_DEFINE} should contain "
                "exactly two argument, "
                f"not {len(operands)} as in '{operands}'"
            )
        macro_key, macro_value = operands
        if not is_variable_name(macro_key):
            raise NetQASMInstrError(f"{macro_key} is not a valid macro key")
        macro_keys.append(macro_key)
    if len(set(macro_keys)) < len(macro_keys):
        raise NetQASMInstrError(f"Macro keys need to be unique, not {macro_keys}")


def _assert_single_preamble_instr(list_of_operands, instr):
    if len(list_of_operands) != 1:
        raise NetQASMInstrError(
            f"Preamble should contain exactly one f{instr} instruction"
        )


def _assert_single_preamble_arg(list_of_operands, instr):
    for operands in list_of_operands:
        if len(operands) != 1:
            raise NetQASMSyntaxError(
                f"Preamble instruction {instr} should contain exactly one argument, "
                f"not {len(operands)} as in '{operands}'"
            )


def _assign_branch_labels(subroutine):
    """Finds assigns the branch labels in a subroutine (inplace)"""
    branch_labels = {}
    command_number = 0
    commands = subroutine.commands
    while command_number < len(commands):
        command = commands[command_number]
        if not isinstance(command, BranchLabel):
            command_number += 1
            continue
        branch_label = command.name
        if branch_label in branch_labels:
            raise NetQASMSyntaxError(
                f"branch labels need to be unique, name {branch_label} already used"
            )
        # Assign the label to the line/command number
        branch_labels[branch_label] = command_number
        # Remove the line
        commands = commands[:command_number] + commands[command_number + 1 :]
    subroutine.commands = commands
    _update_labels(subroutine, branch_labels)


def _update_labels(subroutine, variables: Dict[str, int], from_command=0):
    """Updates labels in a subroutine with given values"""
    for command in subroutine.commands[from_command:]:
        if isinstance(command, ICmd):
            _update_labels_in_command(command, variables)


def _update_labels_in_command(command, variables: Dict[str, int]):
    for i, operand in enumerate(command.operands):
        new_operand = _update_labels_in_operand(operand, variables)
        command.operands[i] = new_operand


def _update_labels_in_operand(operand, labels: Dict[str, int]):
    if isinstance(operand, Label):
        for label, value in labels.items():
            if operand.name == label:
                return value
    return operand


def _get_unused_address(current_addresses):
    for address in count(0):
        if address not in current_addresses:
            return address


def _make_args_operands(subroutine):
    for command in subroutine.commands:
        if not isinstance(command, ICmd):
            continue
        command.operands = command.args + command.operands
        command.args = []


_ALLOW_LABEL_INSTRUCTIONS = [
    GenericInstr.JMP,
    GenericInstr.BEQ,
    GenericInstr.BNE,
    GenericInstr.BEZ,
    GenericInstr.BNZ,
    GenericInstr.BLT,
    GenericInstr.BGE,
]

_REPLACE_CONSTANTS_EXCEPTION = [
    (GenericInstr.SET, 1),
    (GenericInstr.JMP, 0),
    (GenericInstr.BEZ, 1),
    (GenericInstr.BNZ, 1),
    (GenericInstr.BEQ, 2),
    (GenericInstr.BNE, 2),
    (GenericInstr.BLT, 2),
    (GenericInstr.BGE, 2),
    (GenericInstr.BREAKPOINT, 0),
    (GenericInstr.BREAKPOINT, 1),
]

for instr in [GenericInstr.ROT_X, GenericInstr.ROT_Y, GenericInstr.ROT_Z]:
    for index in [1, 2]:
        _REPLACE_CONSTANTS_EXCEPTION.append((instr, index))

for instr in [
    GenericInstr.CROT_X,
    GenericInstr.CROT_Y,
    GenericInstr.CROT_Z,
]:
    for index in [2, 3]:
        _REPLACE_CONSTANTS_EXCEPTION.append((instr, index))

for index in [2, 3, 4, 5]:
    _REPLACE_CONSTANTS_EXCEPTION.append((GenericInstr.MEAS_BASIS, index))


def _replace_constants(commands: List[Union[ICmd, BranchLabel]]):
    current_registers = get_current_registers(commands)

    def reg_and_set_cmd(value, tmp_registers: List[Register], lineno=None):
        for i in range(2**REG_INDEX_BITS):
            register = Register(RegisterName.R, i)
            if str(register) not in current_registers and register not in tmp_registers:
                break
        else:
            raise RuntimeError("Could not replace constant since no registers left")
        set_command = ICmd(
            instruction=GenericInstr.SET,
            args=[],
            operands=[register, value],
            lineno=lineno,
        )
        tmp_registers.append(register)

        return register, set_command

    i = 0
    while i < len(commands):
        command = commands[i]
        if not isinstance(command, ICmd):
            i += 1
            continue
        tmp_registers: List[Register] = []
        for j, operand in enumerate(command.operands):
            if (
                isinstance(operand, int)
                and (command.instruction, j) not in _REPLACE_CONSTANTS_EXCEPTION
            ):
                register, set_command = reg_and_set_cmd(
                    operand, tmp_registers, lineno=command.lineno
                )
                commands.insert(i, set_command)
                command.operands[j] = register

                i += 1
            else:
                if isinstance(operand, ArrayEntry):
                    attrs = ["index"]
                elif isinstance(operand, ArraySlice):
                    attrs = ["start", "stop"]
                else:
                    continue
                for attr in attrs:
                    value = getattr(operand, attr)
                    if isinstance(value, int):
                        register, set_command = reg_and_set_cmd(
                            value, tmp_registers, lineno=command.lineno
                        )
                        commands.insert(i, set_command)
                        setattr(operand, attr, register)

                        i += 1
        i += 1
    return commands


def get_current_registers(commands: List[T_Cmd]) -> Set[str]:
    current_registers = set()
    for command in commands:
        if not isinstance(command, ICmd):
            continue
        for op in command.operands:
            if isinstance(op, Register):
                current_registers.add(str(op))
    return current_registers



from __future__ import annotations

from dataclasses import dataclass
from typing import List, Tuple, Union

from netqasm.lang import encoding
from netqasm.lang.encoding import RegisterName
from netqasm.lang.symbols import Symbols


class Operand:
    pass


@dataclass(eq=True, frozen=True)
class Immediate(Operand):
    value: int

    def __str__(self):
        return str(self.value)


class RegisterMeta:
    @classmethod
    def prefixes(cls) -> List[str]:
        return ["R", "C", "Q", "M"]

    @classmethod
    def parse(cls, name: str) -> Tuple[RegisterName, int]:
        assert len(name) >= 2
        assert name[0] in cls.prefixes()
        group = RegisterName[name[0]]
        index = int(name[1:])
        assert index < 16
        return group, index


@dataclass(eq=True, frozen=True)
class Register(Operand):
    name: RegisterName
    index: int

    def _assert_types(self):
        assert isinstance(self.name, RegisterName)
        assert isinstance(self.index, int)

    def __str__(self):
        return f"{self.name.name}{self.index}"

    @classmethod
    def from_str(cls, name: str) -> Register:
        reg_name, index = RegisterMeta.parse(name)
        return Register(reg_name, index)

    @property
    def cstruct(self):
        self._assert_types()
        return encoding.Register(self.name.value, self.index)

    def __bytes__(self):
        return bytes(self.cstruct)

    @classmethod
    def from_raw(cls, raw: encoding.Register):
        reg_name = RegisterName(raw.register_name)
        return cls(name=reg_name, index=raw.register_index)


@dataclass(eq=True, frozen=True)
class Address(Operand):
    address: Union[int, str]

    def _assert_types(self):
        assert isinstance(self.address, int)

    def __str__(self):
        return f"{Symbols.ADDRESS_START}{self.address}"

    @property
    def cstruct(self):
        self._assert_types()
        return encoding.Address(self.address)

    def __bytes__(self):
        return bytes(self.cstruct)

    @classmethod
    def from_raw(cls, raw: encoding.Address):
        return cls(address=raw.address)


@dataclass
class ArrayEntry(Operand):
    address: Address
    index: Union[Register, int]  # Can ONLY be int when in a "ProtoSubroutine"

    def __post_init__(self):
        if isinstance(self.address, int):
            self.address = Address(self.address)

    def _assert_types(self):
        assert isinstance(self.address, Address)
        assert isinstance(self.index, Register)

    def __str__(self):
        index = f"{Symbols.INDEX_BRACKETS[0]}{self.index}{Symbols.INDEX_BRACKETS[1]}"
        return f"{self.address}{index}"

    @property
    def cstruct(self):
        self._assert_types()
        return encoding.ArrayEntry(
            self.address.cstruct,
            self.index.cstruct,
        )

    def __bytes__(self):
        return bytes(self.cstruct)

    @classmethod
    def from_raw(cls, raw: encoding.ArrayEntry):
        addr = Address.from_raw(raw.address)
        index = Register.from_raw(raw.index)
        entry = cls(address=addr, index=index)
        return entry


@dataclass
class ArraySlice(Operand):
    address: Address
    start: Union[Register, int]  # Can ONLY be int when in a "ProtoSubroutine"
    stop: Union[Register, int]  # Can ONLY be int when in a "ProtoSubroutine"

    def __post_init__(self):
        if isinstance(self.address, int):
            self.address = Address(self.address)

    def _assert_types(self):
        assert isinstance(self.address, Address)
        assert isinstance(self.start, Register)
        assert isinstance(self.stop, Register)

    def __str__(self):
        index = f"{Symbols.INDEX_BRACKETS[0]}{self.start}{Symbols.SLICE_DELIM}{self.stop}{Symbols.INDEX_BRACKETS[1]}"
        return f"{self.address}{index}"

    @property
    def cstruct(self):
        self._assert_types()
        return encoding.ArraySlice(
            self.address.cstruct,
            self.start.cstruct,
            self.stop.cstruct,
        )

    def __bytes__(self):
        return bytes(self.cstruct)

    @classmethod
    def from_raw(cls, raw: encoding.ArraySlice):
        addr = Address.from_raw(raw.address)
        start = Register.from_raw(raw.start)
        stop = Register.from_raw(raw.stop)
        return cls(address=addr, start=start, stop=stop)


@dataclass(eq=True, frozen=True)
class Label:
    name: str

    def _assert_types(self):
        assert isinstance(self.name, str)

    def __str__(self):
        return self.name


@dataclass(eq=True, frozen=True)
class Template(Operand):
    """An operand that does not have a concrete value (it can be filled in later)."""

    name: str

    def _assert_types(self):
        assert isinstance(self.name, str)

    def __str__(self):
        return "{" + self.name + "}"



import pytest

from netqasm.lang.parsing import parse_register
from netqasm.sdk.futures import Future, NonConstantIndexError, NoValueError
from netqasm.sdk.shared_memory import SharedMemory


class MockConnnection:
    def __init__(self):
        self._variables = {}
        self.shared_memory = SharedMemory()


def test_non_constant_index():
    conn = MockConnnection()

    m = Future(conn, address=0, index=parse_register("R0"))
    with pytest.raises(NonConstantIndexError):
        print(m)


def test_no_value():
    conn = MockConnnection()
    conn.shared_memory.init_new_array(address=0, length=1)

    m = Future(conn, address=0, index=0)
    print(m)
    with pytest.raises(NoValueError):
        print(m * 2)


def test_with_value():
    conn = MockConnnection()
    conn.shared_memory.init_new_array(address=0, length=1)
    conn.shared_memory.set_array_part(address=0, index=0, value=4)

    m = Future(conn, address=0, index=0)
    print(m)
    print(m * 2)
    assert m * 2 == 8



from __future__ import annotations

from enum import Enum, auto
from typing import TYPE_CHECKING, Optional, Tuple, Union

from netqasm.sdk.qubit import Qubit

if TYPE_CHECKING:
    from netqasm.sdk import epr_socket as esck
    from netqasm.sdk import futures
    from netqasm.sdk.classical_communication import socket


class _Role(Enum):
    start = auto()
    middle = auto()
    end = auto()


def create_ghz(
    down_epr_socket: Optional[esck.EPRSocket] = None,
    up_epr_socket: Optional[esck.EPRSocket] = None,
    down_socket: Optional[socket.Socket] = None,
    up_socket: Optional[socket.Socket] = None,
    do_corrections: bool = False,
) -> Tuple[Qubit, Union[futures.Future, futures.RegFuture, int]]:
    r"""Local protocol to create a GHZ state between mutliples nodes.

    EPR pairs are generated in a line and turned into a GHZ state by performing half of a Bell measurement.
    That is, CNOT and H are applied but only the control qubit is measured.
    If `do_corrections=False` (default) this measurement outcome is returned along with the qubit to be able to know
    what corrections might need to be applied.
    If the node is at the start or end of the line, the measurement outcome 0 is always returned since there
    is no measurement performed.
    The measurement outcome indicates if the next node in the line should flip its qubit to get the standard
    GHZ state: :math:`|0\rangle^{\otimes n} + |1\rangle^{\otimes n}`.

    On the other hand if `do_corrections=True`, then the classical sockets `down_socket` and/or `up_socket`
    will be used to communicate the outcomes and automatically perform the corrections.

    Depending on if down_epr_socket and/or up_epr_socket is specified the node,
    either takes the role of the:

    * "start", which intialises the process and creates an EPR
      with the next node using the `up_epr_socket`.
    * "middle", which receives an EPR pair on the `down_epr_socket` and then
      creates one on the `up_epr_socket`.
    * "end", which receives an EPR pair on the `down_epr_socket`.

    NOTE There has to be exactly one "start" and exactly one "end" but zero or more "middle".
    NOTE Both `down_epr_socket` and `up_epr_socket` cannot be `None`.

    Parameters
    ----------
    down_epr_socket : :class:`.sdk.epr_socket.esck.EPRSocket`
        The esck.EPRSocket to be used for receiving EPR pairs from downstream.
    up_epr_socket : :class:`.sdk.epr_socket.esck.EPRSocket`
        The esck.EPRSocket to be used for create EPR pairs upstream.
    down_socket : :class:`.sdk.classical_communication.socket.Socket`
        The classical socket to be used for sending corrections, if `do_corrections = True`.
    up_socket : :class:`.sdk.classical_communication.socket.Socket`
        The classical socket to be used for sending corrections, if `do_corrections = True`.
    do_corrections : bool
        If corrections should be applied to make the GHZ in the standard form
        :math:`|0\rangle^{\otimes n} + |1\rangle^{\otimes n}` or not.

    Returns
    -------
    tuple
        Of the form `(q, m)` where `q` is the qubit part of the state and `m` is the measurement outcome.
    """
    if down_epr_socket is None and up_epr_socket is None:
        raise TypeError("Both down_epr_socket and up_epr_socket cannot be None")

    if down_epr_socket is None:
        assert up_epr_socket is not None
        # Start role
        role = _Role.start
        q = up_epr_socket.create_keep()[0]
        assert isinstance(q, Qubit)
        conn = up_epr_socket.conn
        m = 0
    else:
        assert down_epr_socket is not None
        q = down_epr_socket.recv_keep()[0]
        assert isinstance(q, Qubit)
        conn = down_epr_socket.conn
        if up_epr_socket is None:
            # End role
            role = _Role.end
            m = 0
        else:
            # Middle role
            role = _Role.middle
            q_up: Qubit = up_epr_socket.create_keep()[0]  # type: ignore
            # merge the states by doing half a Bell measurement
            q.cnot(q_up)
            m = q_up.measure()

    # Flush the subroutine
    conn.flush()

    if do_corrections:
        if role == _Role.start:
            assert up_socket is not None
            up_socket.send(str(0))
        else:
            assert down_socket is not None
            corr = int(down_socket.recv(maxsize=1))
            if corr == 1:
                q.X()
            if role == _Role.middle:
                assert up_socket is not None
                corr = (corr + m) % 2
                up_socket.send(str(corr))
        m = 0

    return q, m



"""NetQASM application definitions.

NetQASM applications are modeled as pieces of Python code together with metadata.
Generally, applications are multi-node, i.e. they consist of separate chunks of code
that are run by separate nodes.

To distinguish the notion of a multiple-node-spanning collection of code and single-node
piece of code, the following terminology is used:

- A *Program* is code that runs on a single node. It is a Python script whose code is
  executed on (1) the Host component of that node and (2) the quantum node controller
  of that node.
- An *Application* is a collection of Programs (specifically, one Program per node).

"""

import importlib
import os
import sys
from dataclasses import dataclass
from typing import Any, Callable, Dict, List, Optional, Tuple

from netqasm.runtime import env
from netqasm.runtime.interface.config import NetworkConfig, parse_network_config
from netqasm.sdk.config import LogConfig
from netqasm.util.yaml import load_yaml


@dataclass
class Program:
    """Program running on one specific node. Part of a multi-node application.

    :param party: name of the party or role in the multi-node application (protocol).
        E.g. a blind computation application may have two parties: "client" and
        "server". Note that the party name is not necessarily the name of the node this
        Program runs on (which may be, e.g. "Delft").
    :param entry: entry point of the Program. This must be Python function.
    :param args: list of argument names that the entry point expects
    :param results: list of result names that are keys in the dictionary that this
        Program returns on completion
    """

    party: str
    entry: Callable
    args: List[str]
    results: List[str]


@dataclass
class AppMetadata:
    """Metadata about a NetQASM application.

    :param name: name of the application
    :param description: description of the application
    :param authors: list of authors of the application
    :param version: application version
    """

    name: str
    description: str
    authors: List[str]
    version: str


@dataclass
class Application:
    """Static NetQASM application (or protocol) information.

    :param programs: list of programs for each of the parties that are involved
        in this application (or protocol).
    :param metadata: application metadata
    """

    programs: List[Program]
    metadata: Optional[AppMetadata]


@dataclass
class ApplicationInstance:
    """Instantiation of a NetQASM application with concrete input values and
        configuration of the underlying network.

    :param app: static application info
    :param program_inputs: program input values for each of the application's programs
    :param network: configuration for a simulated network
    :param party_alloc: mapping of application parties to nodes in the network
    :param logging_cfg: logging configuration
    """

    app: Application
    program_inputs: Dict[str, Dict[str, Any]]
    network: Optional[NetworkConfig]
    party_alloc: Dict[str, str]
    logging_cfg: Optional[LogConfig]


@dataclass
class ApplicationOutput:
    """Results of a finished run of an ApplicationInstance. Should be subclassed."""

    pass


def load_yaml_file(path: str) -> Any:
    if not os.path.exists(path):
        raise ValueError(f"Could not read file {path} since it does not exist.")
    return load_yaml(path)


def app_instance_from_path(app_dir: str = None) -> ApplicationInstance:
    """
    Create an Application Instance based on files in a directory.
    Uses the current working directory if `app_dir` is None.
    """
    app_dir = os.path.abspath(".") if app_dir is None else os.path.expanduser(app_dir)
    sys.path.append(app_dir)
    program_files = env.load_app_files(app_dir)

    programs = []
    program_inputs = {}
    for party, prog_file in program_files.items():
        prog_module = importlib.import_module(prog_file[: -len(".py")])
        main_func = getattr(prog_module, "main")
        prog = Program(party=party, entry=main_func, args=[], results=[])
        programs += [prog]
        prog_inputs = env.load_app_config_file(app_dir, party)
        program_inputs[party] = prog_inputs

    roles_cfg_path = (
        os.path.abspath(".") if app_dir is None else os.path.join(app_dir, "roles.yaml")
    )
    party_alloc = env.load_roles_config(roles_cfg_path)
    party_alloc = (
        {prog.party: prog.party for prog in programs}
        if party_alloc is None
        else party_alloc
    )

    app = Application(programs=programs, metadata=None)
    app_instance = ApplicationInstance(
        app=app,
        program_inputs=program_inputs,
        network=None,
        party_alloc=party_alloc,
        logging_cfg=None,
    )

    return app_instance


def default_app_instance(programs: List[Tuple[str, Callable]]) -> ApplicationInstance:
    """
    Create an Application Instance with programs that take no arguments.
    """
    program_objects = [
        Program(party=party, entry=entry, args=[], results=[])
        for (party, entry) in programs
    ]
    app = Application(programs=program_objects, metadata=None)
    app_instance = ApplicationInstance(
        app=app,
        program_inputs={party: {} for (party, _) in programs},
        network=None,
        party_alloc={party: party for (party, _) in programs},
        logging_cfg=None,
    )
    return app_instance


def network_cfg_from_path(
    app_dir: str = None, network_config_file: str = None
) -> Optional[NetworkConfig]:
    if network_config_file is None:
        network_config_file = "network.yaml"
    if app_dir is not None:
        network_config_file = os.path.join(app_dir, network_config_file)

    if not os.path.exists(network_config_file):
        return None
    else:
        yaml_dict = load_yaml_file(network_config_file)
        network_cfg = parse_network_config(yaml_dict)
        return network_cfg


def post_function_from_path(
    app_dir: str = None, post_function_file: str = None
) -> Optional[Callable]:
    if post_function_file is None:
        post_function_file = "post_function.yaml"
    if app_dir is not None:
        post_function_file = os.path.join(app_dir, post_function_file)

    if not os.path.exists(post_function_file):
        return None
    else:
        module = importlib.import_module(post_function_file)
        main_func = getattr(module, "main")
        return main_func  # type: ignore



from netqasm.sdk.epr_socket import EprMeasBasis, EPRSocket
from netqasm.sdk.external import NetQASMConnection


def main(app_config=None):
    log_config = app_config.log_config

    epr_socket = EPRSocket("server")

    client = NetQASMConnection(
        app_name=app_config.app_name,
        log_config=log_config,
        epr_sockets=[epr_socket],
        # return_arrays=False,
    )

    num_pairs = 10

    with client:
        outcomes = epr_socket.create_measure(
            number=num_pairs,
            basis_local=EprMeasBasis.Y,
            basis_remote=EprMeasBasis.Y,
        )

    for outcome in outcomes:
        print(f"Client: {outcome.measurement_outcome}")


if __name__ == "__main__":
    main()



from netqasm.sdk.epr_socket import EPRSocket
from netqasm.sdk.external import NetQASMConnection


def main(app_config=None):
    log_config = app_config.log_config

    epr_socket = EPRSocket("client")

    server = NetQASMConnection(
        app_name=app_config.app_name,
        log_config=log_config,
        epr_sockets=[epr_socket],
        # return_arrays=False,
    )

    num_pairs = 10

    with server:
        outcomes = epr_socket.recv_measure(number=num_pairs)

    for outcome in outcomes:
        print(f"Server: {outcome.measurement_outcome}")


if __name__ == "__main__":
    main()



from netqasm.sdk import EPRSocket, Qubit
from netqasm.sdk.external import NetQASMConnection, Socket
from netqasm.sdk.toolbox import set_qubit_state


def main(app_config=None, phi=0.0, theta=0.0):
    # socket for creating an EPR pair with Controller
    controller_epr = EPRSocket("controller")

    # socket for communicating classical messages with Controller
    class_socket = Socket("target", "controller", log_config=app_config.log_config)

    # connect to the back-end
    target = NetQASMConnection(
        app_name=app_config.app_name,
        log_config=app_config.log_config,
        epr_sockets=[controller_epr],
    )

    with target:
        # create one EPR pair with Controller
        epr = controller_epr.recv_keep(1)[0]

        # initialize target qubit of the distributed CNOT
        target_qubit = Qubit(target)
        set_qubit_state(target_qubit, phi, theta)

        # let back-end execute the quantum operations above
        target.flush()

        # wait for Controller's measurement outcome
        m = class_socket.recv()

        # if outcome = 1, apply an X gate on the local EPR half
        if m == "1":
            epr.X()

        # At this point, `epr` is correlated with the control qubit on Controller's side.
        # (If Controller's control was in a superposition, `epr` is now entangled with it.)
        # Use `epr` as the control of a local CNOT on the target qubit.
        epr.cnot(target_qubit)

        # undo any potential entanglement between `epr` and Controller's control qubit
        epr.H()
        epr_meas = epr.measure()
        target.flush()

        # Controller will do a controlled-Z based on the outcome to undo the entanglement
        class_socket.send(str(epr_meas))

        # Wait for an ack before exiting
        assert class_socket.recv() == "ACK"

    return {"epr_meas": int(epr_meas)}



from netqasm.logging.glob import get_netqasm_logger
from netqasm.runtime.application import default_app_instance
from netqasm.sdk import Qubit
from netqasm.sdk.external import NetQASMConnection, simulate_application

logger = get_netqasm_logger()


def run_alice():
    with NetQASMConnection("Alice") as alice:
        count = 0
        num = 10
        for _ in range(num):
            q = Qubit(alice)
            q.rot_X(n=1, d=1)  # pi / 2
            q.rot_X(n=2, d=2)  # 2 pi / 4
            q.rot_Y(n=1, d=1)  # pi / 2
            q.rot_Y(n=2, d=2)  # 2 pi / 4
            m = q.measure()
            alice.flush()
            count += m
        logger.info(count)
        assert count == 0


def test_rotations():
    app_instance = default_app_instance(
        [
            ("Alice", run_alice),
        ]
    )
    simulate_application(app_instance, use_app_config=False, enable_logging=False)



import numpy as np


def bloch_sphere_rep(mat):
    """Computes polar coordinates in Bloch sphere given a single-qubit density matrix

    Parameters
    ----------
    mat : numpy.ndarray
        The single-qubit density matrix

    Returns
    -------
    tuple : (theta, phi, r)
    """
    assert isinstance(mat, np.ndarray), "mat should be instance of numpy.ndarray"
    assert mat.shape == (2, 2), "mat should be a 2x2 matrix"

    # Compute cartesian coordinates
    ax = mat[0, 1] + mat[1, 0]
    ay = (mat[1, 0] - mat[0, 1]) / 1j
    az = 2 * mat[0, 0] - 1

    # Compute polar coordinates
    r = np.linalg.norm([ax, ay, az])
    assert r < 1.001, "mat not normalized"
    if r == 0:
        return (0, 0, 0)
    theta = np.arccos(az / r)
    if theta in [0, np.pi]:
        return (theta, 0, r)
    phi = np.arccos(ax / r / np.sin(theta))
    if ay < 0:
        phi += np.pi

    return (theta, phi, r)
